@@ -1,4 +1,4 @@
-/* $Id: display.c,v 1.3 2000/01/21 07:09:16 aito Exp $ */
+/* $Id: display.c,v 1.5 2000/04/06 01:24:59 aito Exp $ */
 #include <signal.h>
 #include "fm.h"
 
@@ -83,7 +83,7 @@ static void EFFECT_ACTIVE_END
 }
     
 
-#else /* no color */
+#else /* not COLOR */
 
 #define EFFECT_ANCHOR_START       underline()
 #define EFFECT_ANCHOR_END         underlineend()
@@ -94,12 +94,12 @@ static void EFFECT_ACTIVE_END
 #define EFFECT_ACTIVE_START       bold()
 #define EFFECT_ACTIVE_END         boldend()
 
-#endif /* COLOR */
+#endif /* not COLOR */
 
 
 #ifndef KANJI_SYMBOLS
 static char g_rule[] = "ntwluxkavmqajaaa";
-#endif
+#endif /* not KANJI_SYMBOLS */
 
 /*
  * Terminate routine.
@@ -115,7 +115,7 @@ fmTerm(void)
 #ifdef MOUSE
     if (use_mouse)
       mouse_end();
-#endif
+#endif /* MOUSE */
     reset_tty();
     fmInitialized = NIL;
   }
@@ -149,7 +149,7 @@ fmInit(void)
 #ifdef MOUSE
     if (use_mouse)
       mouse_init();
-#endif
+#endif /* MOUSE */
   }
   fmInitialized = T;
 }
@@ -164,7 +164,7 @@ static int     ulmode = 0, somode = 0, bomode = 0;
 static int     anch_mode = 0, emph_mode = 0, imag_mode = 0, form_mode = 0, active_mode = 0;
 #ifndef KANJI_SYMBOLS
 static int     graph_mode = 0;
-#endif
+#endif /* not KANJI_SYMBOLS */
 
 void
 displayBuffer(Buffer * buf, int mode)
@@ -176,7 +176,7 @@ displayBuffer(Buffer * buf, int mode)
     buf->width = COLS;
   if (buf->height == 0)
     buf->height = LASTLINE+1;
-  if (buf->width != COLS - 5 && buf->type && !strcmp(buf->type, "text/html"))
+  if (buf->width != INIT_BUFFER_WIDTH && buf->type && !strcmp(buf->type, "text/html"))
     reshapeBuffer(buf);
   if (mode == B_FORCE_REDRAW ||
       mode == B_SCROLL ||
@@ -219,12 +219,12 @@ displayBuffer(Buffer * buf, int mode)
   else
 #if LANG == JA
   msg = Sprintf("㢬Viewing <%s> ", cleanup_str(buf->buffername));
-#else
+#else /* LANG != JA */
   msg = Sprintf("<=UpDn Viewing <%s> ", cleanup_str(buf->buffername));
-#endif
-#else
+#endif /* LANG != JA */
+#else /* not MOUSE */
   msg = Sprintf("Viewing <%s> ", cleanup_str(buf->buffername));
-#endif
+#endif /* not MOUSE */
   if (displayLink) {
     aa = retrieveCurrentAnchor(buf);
     if (aa) {
@@ -238,9 +238,9 @@ displayBuffer(Buffer * buf, int mode)
         Strcat(msg,Strsubstr(s, 0, (l - 2) / 2));
 #if LANG == JA
         Strcat_charp(msg,"");
-#else
+#else /* LANG != JA */
         Strcat_charp(msg,"..");
-#endif
+#endif /* LANG != JA */
         l = buf->width - msg->length;
         Strcat(msg,Strsubstr(s, s->length - l, l));
       } else {
@@ -270,9 +270,9 @@ redrawNLine(Buffer * buf, int n)
     EFFECT_ANCHOR_END_C;
 #ifdef BG_COLOR
     setbcolor(bg_color);
-#endif
+#endif /* BG_COLOR */
   }
-#endif
+#endif /* COLOR */
   for (i = 0, l = buf->topLine; i < LASTLINE; i++) {
     if (i >= LASTLINE-n || i < -n)
       l0 = redrawLine(buf, l, i);
@@ -322,7 +322,7 @@ redrawLine(Buffer * buf, Line * l, int i)
     addChar(' ', 0);
     rcol++;
   } else 
-#endif
+#endif /* JP_CHARSET */
   if (*p == '\t') {
     do {
       addChar(' ', *pr);
@@ -339,7 +339,7 @@ redrawLine(Buffer * buf, Line * l, int i)
 #ifdef JP_CHARSET
    if (rcol-column == COLS-1 && CharType(pr[j]) == PC_KANJI1)
       break;
-#endif
+#endif /* JP_CHARSET */
    if (p[j] == '\t') {
       do {
 	addChar(' ', 0);
@@ -388,7 +388,7 @@ redrawLine(Buffer * buf, Line * l, int i)
     graph_mode = NIL;
     graphend();
   }
-#endif
+#endif /* not KANJI_SYMBOLS */
   if (rcol - column < COLS)
     clrtoeolx();
   return l->next;
@@ -427,7 +427,7 @@ redrawLineRegion(Buffer * buf, Line * l, int i, int bpos, int epos)
       }
       rcol++;
     } else 
-#endif
+#endif /* JP_CHARSET */
     if (p[j] == '\t') {
       do {
 	if (j >= bcol && j < ecol) {
@@ -481,7 +481,7 @@ redrawLineRegion(Buffer * buf, Line * l, int i, int bpos, int epos)
     graph_mode = NIL;
     graphend();
   }
-#endif
+#endif /* not KANJI_SYMBOLS */
  
   return rcol - column;
 }
@@ -517,7 +517,7 @@ do_effects(Lineprop m)
     graphend();
     graph_mode = NIL;
   }
-#endif
+#endif /* not KANJI_SYMBOLS */
 
   /* effect start */
   do_effect1(PE_UNDER, ulmode, underline(), underlineend())
@@ -535,7 +535,7 @@ do_effects(Lineprop m)
       graph_mode = T;
     }
   }
-#endif
+#endif /* not KANJI_SYMBOLS */
 }
 
 void
@@ -549,14 +549,14 @@ addChar(char c, Lineprop mode)
   if (c & 0x80 && CharType(mode) == PC_ASCII) {
     addch(' ');
   } else 
-#else
+#else /* not JP_CHARSET */
   do_effects(m);
-#endif
+#endif /* not JP_CHARSET */
 #ifndef KANJI_SYMBOLS
   if (graph_mode)
     addch(g_rule[PC_RULE_VAL(m)]);
   else
-#endif
+#endif /* not KANJI_SYMBOLS */
   if (c == '\t') {
     addch(c);
   } else if (!(c & ~0x1f) && c != CTRL_J) {	/* Control code */
@@ -669,7 +669,7 @@ cursorRight(Buffer * buf)
 #ifdef JP_CHARSET
   if (CharType(p[i]) == PC_KANJI1)
     delta = 2;
-#endif
+#endif /* JP_CHARSET */
   if (i >= buf->currentLine->len - delta)
     return;
   if (buf->visualpos >= COLS - delta)
@@ -693,7 +693,7 @@ cursorLeft(Buffer * buf)
 #ifdef JP_CHARSET
   if (i >=2 && CharType(p[i - 1]) == PC_KANJI2)
     delta = 2;
-#endif
+#endif /* JP_CHARSET */
   if (i > delta)
     buf->pos = i - delta;
   else
