@@ -1,4 +1,4 @@
-/* $Id: etc.c,v 1.8 2000/01/14 06:13:27 aito Exp $ */
+/* $Id: etc.c,v 1.16 2000/04/06 01:24:59 aito Exp $ */
 #include "fm.h"
 #include <pwd.h>
 #include "myctype.h"
@@ -9,11 +9,21 @@
 #ifdef GETCWD
 #include <unistd.h>
 #include <sys/param.h>
-#endif
+#endif /* GETCWD */
+
+#ifdef __EMX__
+#include <process.h>
+#endif /* __EMX__ */
 
 #include <sys/types.h>
 #include <time.h>
 #include <sys/wait.h>
+#include <signal.h>
+
+#ifdef	__WATT32__
+#define	read(a,b,c)	read_s(a,b,c)
+#define	close(x)	close_s(x)
+#endif /* __WATT32__ */
 
 #ifndef STRCASECMP
 int
@@ -48,7 +58,7 @@ strncasecmp(char *s1, char *s2, int n)
     return x;
   return 0;
 }
-#endif
+#endif /* not STRCASECMP */
 
 int
 arg_is(char *str, char *tag)
@@ -263,7 +273,7 @@ checkType(char *istr, Lineprop * oprop, int len)
     if (ShowEffect
 #ifdef JP_CHARSET
 	&& mode != PC_KANJI1
-#endif
+#endif /* JP_CHARSET */
 	) {
       if (!strncmp(str, "_\b", 2)) {
 	str += 2;
@@ -280,7 +290,7 @@ checkType(char *istr, Lineprop * oprop, int len)
 	  *(prop - 1) |= PE_UNDER;
 	  *(prop - 2) |= PE_UNDER;
 	} else
-#endif
+#endif /* JP_CHARSET */
 	if (s->length > 0) {
 	  str += 2;
 	  *(prop - 1) |= PE_UNDER;
@@ -295,7 +305,7 @@ checkType(char *istr, Lineprop * oprop, int len)
 	  *(prop - 1) |= PE_UNDER;
 	  *(prop - 2) |= PE_UNDER;
 	} else
-#endif
+#endif /* JP_CHARSET */
 	if (s->length > 0) {
 	  str += 3;
 	  *(prop - 1) |= PE_UNDER;
@@ -316,7 +326,7 @@ checkType(char *istr, Lineprop * oprop, int len)
 	    str += 2;
 	  }
 	} else
-#endif
+#endif /* JP_CHARSET */
 	if (s->length > 0) {
 	  if (*(str-1) == *(str+2)) {
 	    *(prop - 1) |= PE_BOLD;
@@ -343,7 +353,7 @@ checkType(char *istr, Lineprop * oprop, int len)
 	    str++;
 	  }
 	} else
-#endif
+#endif /* JP_CHARSET */
 	if (s->length > 0) {
 	  if (*(str-1) == *(str+1)) {
 	    *(prop - 1) |= PE_BOLD;
@@ -370,12 +380,12 @@ checkType(char *istr, Lineprop * oprop, int len)
     if (mode == PC_KANJI2)
       *(prop++) = (leffect | mode);
     else
-#endif
+#endif /* JP_CHARSET */
       *(prop++) = (effect | mode);
     Strcat_char(s,*(str++));
 #ifdef JP_CHARSET
     if (mode != PC_KANJI1)
-#endif
+#endif /* JP_CHARSET */
       effect = PE_NORMAL;
   }
   return s;
@@ -446,7 +456,7 @@ bzero(void *ptr, int len)
   for (i = 0; i < len; i++)
     *(ptr++) = 0;
 }
-#endif
+#endif /* NOBCOPY */
 
 /* fgets() substitute which terminates \n, \r or \n\r */
 char           *
@@ -485,7 +495,7 @@ Strmyfgets(FILE * stream)
 
   if (feof(stream))
     return p;
-  while (c = getc(stream), !feof(stream)) {
+  while (c = getc(stream), !feof(stream) && !ferror(stream)) {
     if (c == '\0') {
       Strcat_charp(p,"^@");
       continue;
@@ -512,7 +522,7 @@ Strkconv(Str s, int fcode, int tcode)
     return s;
   return conv(s->ptr,fcode,tcode);
 }
-#endif
+#endif /* JP_CHARSET */
 
 char*
 mybasename(char* s)
@@ -534,7 +544,7 @@ strerror(int errno)
   extern char *sys_errlist[];
   return sys_errlist[errno];
 }
-#endif
+#endif /* not STRERROR */
 
 #ifndef SYS_ERRLIST
 char **sys_errlist;
@@ -552,7 +562,7 @@ prepare_sys_errlist()
   for (i = 1; i < n; i++)
     sys_errlist[i] = strerror(i);
 }
-#endif
+#endif /* not SYS_ERRLIST */
 
 int
 next_status(char c, int *status)
@@ -620,7 +630,7 @@ next_status(char c, int *status)
       *status = R_ST_NORMAL;
       break;
     default:
-      *status = R_ST_TAG;
+      *status = R_ST_IRRTAG;
     }
     return 0;
   case R_ST_CMNT2:
@@ -632,7 +642,7 @@ next_status(char c, int *status)
       *status = R_ST_NORMAL;
       break;
     default:
-      *status = R_ST_TAG;
+      *status = R_ST_IRRTAG;
     }
     return 0;
   case R_ST_CMNT:
@@ -677,6 +687,10 @@ next_status(char c, int *status)
       break;
     }
     return 0;
+  case R_ST_IRRTAG:
+    if (c == '>')
+      *status = R_ST_NORMAL;
+    return 0;
   }
   /* notreached */
   return 0;
@@ -701,7 +715,8 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
         p--;
         break;
       }
-      if (prev_status == R_ST_NCMNT2 || prev_status == R_ST_NCMNT3) {
+      if (prev_status == R_ST_NCMNT2 || prev_status == R_ST_NCMNT3 ||
+	  prev_status == R_ST_IRRTAG) {
 	p++;
 	goto proc_end;
       }
@@ -755,6 +770,7 @@ read_token(Str buf, char **instr, int *status, int pre, int append)
       Strcat_char(buf, *p);
       break;
     case R_ST_CMNT:
+    case R_ST_IRRTAG:
       if (!append)
         Strclear(buf);
       break;
@@ -882,7 +898,7 @@ get_ctype(unsigned char ch, Lineprop prev_mode)
     if (is_kanji1(ch, InnerCode))
 	return PC_KANJI1;
     else 
-#endif
+#endif /* JP_CHARSET */
     if (!(ch & ~0x1f))
 	return PC_CTRL;
     else
@@ -1068,7 +1084,7 @@ StrmySSgets(SSLStream ss)
   }
   return p;
 }
-#endif
+#endif /* USE_SSL */
 
 static Str
 sgets(char **str)
@@ -1109,11 +1125,11 @@ StrUFgets(URLFile * f)
     if (!f->stream.ss || f->stream.ss->status <= 0)
       f->iseof = T;
     break;
-#endif
+#endif /* USE_SSL */
   case SMT_FILE:
   default:
     s = Strfgets(f->stream.f);
-    if (!f->stream.f || feof(f->stream.f))
+    if (!f->stream.f || feof(f->stream.f) || ferror(f->stream.f))
       f->iseof = T;
     break;
   }
@@ -1138,11 +1154,11 @@ StrmyUFgets(URLFile * f)
     if (!f->stream.ss || f->stream.ss->status <= 0)
       f->iseof = T;
     break;
-#endif
+#endif /* USE_SSL */
   case SMT_FILE:
   default:
     s = Strmyfgets(f->stream.f);
-    if (!f->stream.f || feof(f->stream.f))
+    if (!f->stream.f || feof(f->stream.f) || ferror(f->stream.f))
       f->iseof = T;
     break;
   }
@@ -1168,12 +1184,12 @@ UFgetc(URLFile *f)
     if (!f->stream.ss || f->stream.ss->status <= 0)
       f->iseof = T;
     break;
-#endif
+#endif /* USE_SSL */
   case SMT_FILE:
   default:
     if (f->stream.f)
       _prev_c = getc(f->stream.f);
-    if (!f->stream.f || feof(f->stream.f))
+    if (!f->stream.f || feof(f->stream.f) || ferror(f->stream.f))
       f->iseof = T;
     break;
   }
@@ -1193,7 +1209,7 @@ UFundogetc(URLFile *f)
     if (f->stream.ss)
       SSundogetc(f->stream.ss);
     break;
-#endif
+#endif /* USE_SSL */
   case SMT_FILE:
   default:
     if (f->stream.f)
@@ -1214,7 +1230,7 @@ UFclose0(URLFile *f, void (*closep)())
   case SMT_SSL:
     SSclose(f->stream.ss);
     break;
-#endif
+#endif /* USE_SSL */
   case SMT_FILE:
   default:
     if (closep)
@@ -1229,50 +1245,27 @@ UFclose(URLFile *f)
   UFclose0(f, f->close);
 }
 
+Str
+quoteShell(char *string)
+{
+  Str str = Strnew();
+
+  while (*string != '\0') {
+    if (!IS_ALNUM(*string) && *string != '_' && *string != '.' &&
+	*string != ':' && *string != '/')
+      Strcat_char(str, '\\');
+    Strcat_char(str, *(string++));
+  }
+  return str;
+}
+
 void
 mySystem(char *command,int background)
 {
-  char *file,*p,*q;
-  char **argv;
-  int argc;
-  int argv_size;
-  pid_t pid;
-
-  argv_size = 10;
-  argv = New_N(char*,argv_size);
-  argc = 0;
-
-  /* get filename */
-  for (p = command; *p && !IS_SPACE(*p); p++);
-  file = argv[argc++] = allocStr(command,p-command);
-
-  for (;;) {
-    for (; *p && IS_SPACE(*p); p++);
-    if (*p == '\0') break;
-    for (q = p; *q && !IS_SPACE(*q); q++);
-    if (argc == argv_size-1) {
-      argv_size *= 2;
-      argv = New_Reuse(char*,argv,argv_size);
-    }
-    if (*p == '\'' && *(q-1) == '\'') {
-      argv[argc++] = allocStr(p+1,q-p-2);
-    } else {
-      argv[argc++] = allocStr(p,q-p);
-    }
-    p = q;
-  }
-  argv[argc] = NULL;
-  if ((pid = fork()) == 0) {
-    /* child */
-    if (background)
-      setsid();
-    if (execvp(file,argv)) {
-      /* error */
-      exit(1);
-    }
-  }
-  if (!background)
-    waitpid(pid,NULL,0);
+  Str cmd = Strnew_charp(command);
+  if (background)
+    Strcat_char(cmd,'&');
+  system(cmd->ptr);
 }
     
 char           *
@@ -1448,7 +1441,7 @@ mymktime(char *timestr)
 
 #ifdef DEBUG
   fprintf(stderr, "mktime: %s\n", timestr);
-#endif
+#endif /* DEBUG */
   
   while (*s && isalpha((unsigned char)*s))
     s++;
@@ -1504,7 +1497,7 @@ mymktime(char *timestr)
 #ifdef DEBUG
   fprintf(stderr, "year=%d month=%d day=%d hour:min:sec=%d:%d:%d\n",
 	  year, mon, day, hour, min, sec);
-#endif
+#endif /* DEBUG */
 
   mon -= 3;
   if (mon < 0) {
@@ -1529,7 +1522,7 @@ FQDN(char *host)
   char *p;
 #ifndef INET6
   struct hostent *entry;
-#else
+#else /* INET6 */
   int *af;
 #endif /* INET6 */
 
@@ -1546,7 +1539,7 @@ FQDN(char *host)
     return NULL;
 
   return entry->h_name;
-#else
+#else /* INET6 */
   for (af = ai_family_order_table[DNS_order]; ; af++) {
       int error;
       struct addrinfo hints;
@@ -1580,7 +1573,7 @@ FQDN(char *host)
   }
   /* all failed */
   return NULL;
-#endif
+#endif /* INET6 */
 }
 
-#endif
+#endif /* USE_COOKIE */
