@@ -7,12 +7,11 @@
 
 #include "fm.h"
 #include "html.h"
-#include "pq.h"
 #include "myctype.h"
 
 #ifdef DEBUG
 #include <malloc.h>
-#endif
+#endif /* DEBUG */
 
 typedef struct _FTP {
   FILE *rcontrol;
@@ -104,8 +103,8 @@ FtpLogin(FTP *ftp_return, char *host, char *user, char *pass)
   fd = openSocket(host,"ftp",21);
   if (fd < 0)
     return -1;
-  ftp->rcontrol = fdopen(fd,"r");
-  ftp->wcontrol = fdopen(dup(fd),"w");
+  ftp->rcontrol = fdopen(fd,"rb");
+  ftp->wcontrol = fdopen(dup(fd),"wb");
   ftp->data = NULL;
   tmp = read_response(ftp);
   if (atoi(tmp->ptr) != 220)
@@ -158,7 +157,7 @@ ftp_pasv(FTP ftp)
   data_s = openSocket(tmp->ptr,"",p1*256+p2);
   if (data_s < 0)
     return -1;
-  ftp->data = fdopen(data_s,"r");
+  ftp->data = fdopen(data_s,"rb");
   return 0;
 }
 
@@ -341,7 +340,8 @@ openFTP(ParsedURL * pu)
   char *pass;
   char *fn;
   char *qdir;
-  PriorityQueue *pq;
+  char **flist;
+  int i,nfile,nfile_max = 100;
   Str pwd;
   int add_auth_cookie_flag;
   char *realpath = NULL;
@@ -403,7 +403,7 @@ openFTP(ParsedURL * pu)
       pathStr = conv(realpath, code, InnerCode);
       realpath = pathStr->ptr;
    }
-#endif
+#endif /* JP_CHARSET */
     pu->file = realpath;
     return FTPDATA(ftp);
   }
@@ -442,8 +442,7 @@ ftp_dir:
   }
   host = Strnew_charp("ftp://");
   if(pu->user){
-    Strcat_charp(host,pu->user);
-    Strcat_charp(host,"@");
+    Strcat_m_charp(host,pu->user,"@",NULL);
   }
   Strcat_charp(host,pu->host);
   if (Strlastchar(host) == '/')
@@ -464,14 +463,15 @@ ftp_dir:
   } else {
     Strcat_charp(FTPDIRtmp, "<ul><li><a href=\"");
   }
-  Strcat(FTPDIRtmp,host);
-  Strcat_charp(FTPDIRtmp,htmlquote_str(tmp2->ptr));
-  Strcat_charp(FTPDIRtmp,"\">[Upper Directory]</a>\n");
+  Strcat_m_charp(FTPDIRtmp,host->ptr,
+		 htmlquote_str(tmp2->ptr),
+		 "\">[Upper Directory]</a>\n",NULL);
 
-  pq = PQ_new(100,string_alphabetical);
+  flist = New_N(char*,nfile_max);
+  nfile = 0;
   if (sv_type == UNIXLIKE_SERVER) {
     char *name, *date, *size, *type_str;
-    int ftype, max_len, len, i;
+    int ftype, max_len, len, i, j;
     Str line_tmp;
 
     max_len = 0;
@@ -497,9 +497,15 @@ ftp_dir:
       }
       if (max_len < len) max_len = len;
       line_tmp = Sprintf("%s%s %-12.12s %6.6s", name, type_str, date, size);
-      PQ_push(pq,(void*)line_tmp->ptr);
+      flist[nfile++] = line_tmp->ptr;
+      if (nfile == nfile_max) {
+	nfile_max *= 2;
+	flist = New_Reuse(char*,flist,nfile_max);
+      }
     }
-    while ((fn = (char*)PQ_pop(pq)) != NULL) {
+    qsort(flist,nfile,sizeof(char*),strCmp);
+    for (j = 0; j < nfile; j++) {
+      fn = flist[j];
       date = fn + strlen(fn) - 20;
       if (*(date - 1) == '/') {
         ftype = FTPDIR_DIR;
@@ -513,12 +519,12 @@ ftp_dir:
       }
       date++;
       len = strlen(fn);
-      Strcat_charp(FTPDIRtmp,"<a href=\"");
-      Strcat(FTPDIRtmp,host);
-      Strcat_charp(FTPDIRtmp,qdir);
-      Strcat_charp(FTPDIRtmp,ftp_escape_str(fn));
-      Strcat_charp(FTPDIRtmp,"\">");
-      Strcat_charp(FTPDIRtmp,htmlquote_str(fn));
+      Strcat_m_charp(FTPDIRtmp,"<a href=\"",
+		     host->ptr,
+		     qdir,
+		     ftp_escape_str(fn),
+		     "\">",
+		     htmlquote_str(fn),NULL);
       if (ftype == FTPDIR_DIR) {
         Strcat_charp(FTPDIRtmp,"/");
         len++;
@@ -534,24 +540,27 @@ ftp_dir:
           Strcat_charp(FTPDIRtmp," ");
         }
       }
-      Strcat_charp(FTPDIRtmp,date);
-      Strcat_charp(FTPDIRtmp,"\n");
+      Strcat_m_charp(FTPDIRtmp,date,"\n",NULL);
     }
     Strcat_charp(FTPDIRtmp, "</pre></body></html>\n");
   } else {
     while (tmp2 = Strfgets(FTPDATA(ftp)),tmp2->length > 0) {
       Strchop(tmp2);
-      fn = mybasename(tmp2->ptr);
-      PQ_push(pq,(void*)fn);
+      flist[nfile++] = mybasename(tmp2->ptr);
+      if (nfile == nfile_max) {
+	nfile_max *= 2;
+	flist = New_Reuse(char*,flist,nfile_max);
+      }
     }
-    while ((fn = (char*)PQ_pop(pq)) != NULL) {
-      Strcat_charp(FTPDIRtmp,"<li><a href=\"");
-      Strcat(FTPDIRtmp,host);
-      Strcat_charp(FTPDIRtmp,qdir);
-      Strcat_charp(FTPDIRtmp,ftp_escape_str(fn));
-      Strcat_charp(FTPDIRtmp,"\">");
-      Strcat_charp(FTPDIRtmp,htmlquote_str(fn));
-      Strcat_charp(FTPDIRtmp,"</a>\n");
+    qsort(flist,nfile,sizeof(char*),strCmp);
+    for (i = 0; i < nfile; i++) {
+      fn = flist[i];
+      Strcat_m_charp(FTPDIRtmp,"<li><a href=\"",
+		     host->ptr,qdir,
+		     ftp_escape_str(fn),
+		     "\">",
+		     htmlquote_str(fn),
+		     "</a>\n", NULL);
     }
     Strcat_charp(FTPDIRtmp, "</ul></body></html>\n");
   }
