@@ -14,6 +14,4 @@ typedef struct _FuncList {
 	void (*func)();
 } FuncList;
 
-FuncList *getFuncList(char*,FuncList*,int);
-
-#endif
+#endif /* not FUNC_H */
