@@ -3,8 +3,11 @@
 #include "gc.h"
 #include "Str.h"
 
+extern char * conv_latin1( int ch );
+extern int getescapechar ( char **s );
 extern char * getescapecmd ( char **s );
 extern char * allocStr ( char *s, int len );
+extern int strCmp ( const void *s1, const void *s2 );
 extern void copydicname ( char *s, char *fn );
 extern char * currentdir ( void );
 extern char * cleanupName ( char *name );
@@ -24,4 +27,4 @@ extern char * expandPath ( char *name );
 #define NewAtom_N(type,n)	((type*)GC_MALLOC_ATOMIC((n)*sizeof(type)))
 #define New_Reuse(type,ptr,n)   ((type*)GC_REALLOC((ptr),(n)*sizeof(type)))
 
-#endif
+#endif /* INDEP_H */
