@@ -1,5 +1,5 @@
 /* Marker extension by okabe */
-/* $Id: main.c,v 1.10 2000/01/21 07:18:57 aito Exp $ */
+/* $Id: main.c,v 1.21 2000/04/06 01:25:00 aito Exp $ */
 #define MAINPROGRAM
 #include "fm.h"
 #include <signal.h>
@@ -9,14 +9,16 @@
 #include "myctype.h"
 #ifdef USE_GPM
 #include <gpm.h>
+#endif /* USE_GPM */
+#if defined(USE_GPM) || defined(USE_SYSMOUSE)
 extern int do_getch();
 #define getch()	do_getch()
-#endif /* USE_GPM */
+#endif /* defined(USE_GPM) || defined(USE_SYSMOUSE) */
 
-extern void (*GlobalKeymap[])();
-extern void (*EscKeymap[])();
-extern void (*EscBKeymap[])();
-extern void (*EscDKeymap[])();
+#ifdef __EMX__
+#include "os2emx.h"
+#define ctrl(c) ((c)&0x1F)
+#endif /*  __EMX__ */
 
 #define DSTR_LEN	256
 
@@ -30,7 +32,7 @@ Hist    *ShellHist;
 
 #define N_EVENT_QUEUE 10
 typedef struct {
-  void (*cmd)();
+  int cmd;
   void *user_data;
 } Event;
 static void *cmd_argument;
@@ -48,7 +50,7 @@ static void      cmd_loadBuffer(Buffer *buf, int prop, int link);
 static void      keyPressEventProc(int c);
 #ifdef USE_MARK
 static void cmd_mark(Lineprop * p);
-#endif
+#endif /* USE_MARK */
 
 static int display_ok = NIL;
 int w3m_dump = 0;
@@ -57,32 +59,12 @@ int w3m_dump_head = 0;
 static void dump_source(Buffer*);
 static void dump_head(Buffer*);
 int prec_num = 0;
+int on_target = 1;
 
 #define PREC_NUM (prec_num ? prec_num : 1)
 #define PREC_LIMIT 10000
 
-#if defined(AIX) || defined(linux)
-/* to cope with Boehm GC... */
-
-#define MAIN real_main
-
-#if defined(DEBIAN)
-# include "gc/private/gc_priv.h"
-#else
-# include "gc_private.h"
-#endif
-int real_main(int,char**,char**);
-
-int
-main(int argc, char **argv, char **envp)
-{
-    int dummy;
-    GC_stackbottom = (ptr_t)(&dummy);
-    return(real_main(argc, argv, envp));
-}
-#else
-#define MAIN main
-#endif
+#include "gcmain.c"
 
 static void
 usage()
@@ -96,7 +78,7 @@ usage()
   fprintf(stderr, "    -s               Shift_JIS\n");
   fprintf(stderr, "    -j               JIS\n");
   fprintf(stderr, "    -e               EUC-JP\n");
-#endif
+#endif /* JP_CHARSET */
   fprintf(stderr, "    -B               load bookmark\n");
   fprintf(stderr, "    -bookmark file   specify bookmark file\n");
   fprintf(stderr, "    -T type          specify content-type\n");
@@ -104,7 +86,7 @@ usage()
   fprintf(stderr, "    -v               visual startup mode\n");
 #ifdef COLOR
   fprintf(stderr, "    -M               monochrome display\n");
-#endif
+#endif /* COLOR */
   fprintf(stderr, "    -F               automatically render frame\n");
   fprintf(stderr, "    -dump            dump formatted page into stdout\n");
   fprintf(stderr, "    -cols width      specify column width (used with -dump)\n");
@@ -115,13 +97,13 @@ usage()
   fprintf(stderr, "    -no-proxy        don't use proxy\n");
 #ifdef MOUSE
   fprintf(stderr, "    -no-mouse        don't use mouse\n");
-#endif
+#endif /* MOUSE */
 #ifdef USE_COOKIE
   fprintf(stderr, "    -cookie          use cookie (-no-cookie: don't use cookie)\n");
-#endif
+#endif /* USE_COOKIE */
 #ifndef KANJI_SYMBOLS
   fprintf(stderr, "    -no-graph        don't use graphic character\n");
-#endif
+#endif /* not KANJI_SYMBOLS */
   fprintf(stderr, "    -S               squeeze multiple blank lines\n");
   fprintf(stderr, "    -W               toggle wrap search mode\n");
   fprintf(stderr, "    -X               don't use termcap init/deinit\n");
@@ -147,7 +129,7 @@ MAIN(int argc, char **argv, char **envp)
 
 #ifndef SYS_ERRLIST
   prepare_sys_errlist();
-#endif
+#endif /* not SYS_ERRLIST */
 
   NO_proxy_domains = newTextList();
   fileToDelete = newTextList();
@@ -157,6 +139,9 @@ MAIN(int argc, char **argv, char **envp)
 
   BookmarkFile = NULL;
   rc_dir = expandName(RC_DIR);
+  i = strlen(rc_dir);
+  if (i > 1 && rc_dir[i] == '/')
+    rc_dir[i] = '\0';
   config_file = rcFile("config");
 
   /* argument search 1 */
@@ -175,10 +160,10 @@ MAIN(int argc, char **argv, char **envp)
   initKeymap();
 #ifdef MENU
   initMenu();
-#endif
+#endif /* MENU */
 #ifdef USE_COOKIE
   initCookie();
-#endif
+#endif /* USE_COOKIE */
 
   SearchHist = initHist(HIST_SIZE + 1);
   MarkHist   = initHist(HIST_SIZE + 1);
@@ -188,9 +173,9 @@ MAIN(int argc, char **argv, char **envp)
 #ifdef USE_HISTORY
   URLHist    = initHist(URLHistSize + 1);
   loadHistory(URLHist);
-#else
+#else /* not USE_HISTORY */
   URLHist    = initHist(HIST_SIZE + 1);
-#endif
+#endif /* not USE_HISTORY */
 
   if (HTTP_proxy == NULL &&
       ((p = getenv("HTTP_PROXY")) ||
@@ -242,11 +227,11 @@ MAIN(int argc, char **argv, char **envp)
 	DisplayCode = CODE_JIS1;
       else if (!strcmp("-e", argv[i]))
 	DisplayCode = CODE_EUC;
-#endif
+#endif /* JP_CHARSET */
 #ifndef KANJI_SYMBOLS
       else if (!strcmp("-no-graph",argv[i]))
         no_graphic_char = T;
-#endif
+#endif /* not KANJI_SYMBOLS */
       else if (!strcmp("-T", argv[i]))
 	DefaultType = argv[++i];
       else if (!strcmp("-m", argv[i]))
@@ -256,7 +241,7 @@ MAIN(int argc, char **argv, char **envp)
 #ifdef COLOR
       else if (!strcmp("-M", argv[i]))
 	useColor = NIL;
-#endif     
+#endif /* COLOR */     
       else if (!strcmp("-B", argv[i]))
 	load_bookmark = T;
       else if (!strcmp("-bookmark",argv[i]))
@@ -305,7 +290,7 @@ MAIN(int argc, char **argv, char **envp)
 	mouse_end();
 	use_mouse = NIL;
       }
-#endif
+#endif /* MOUSE */
 #ifdef USE_COOKIE
       else if (!strcmp("-no-cookie", argv[i])) {
 	use_cookie = NIL;
@@ -313,7 +298,7 @@ MAIN(int argc, char **argv, char **envp)
       else if (!strcmp("-cookie", argv[i])) {
 	use_cookie = T;
       }
-#endif
+#endif /* USE_COOKIE */
       else if (!strcmp("-S", argv[i]))
 	squeezeBlankLine = T;
       else if (!strcmp("-X", argv[i]))
@@ -334,25 +319,82 @@ MAIN(int argc, char **argv, char **envp)
     }
   }
 
+#ifdef	__WATT32__
+  if (w3m_debug)
+     dbug_init();
+  sock_init();
+#endif
+
   Firstbuf = NULL;
   Currentbuf = NULL;
   if (BookmarkFile == NULL)
     BookmarkFile = rcFile(BOOKMARK);
-  if (load_bookmark || load_argc > 0) {
-    /* if any arguments are given... */
-    if (!w3m_dump && !w3m_halfdump)
-      fmInit();
-    if (w3m_halfdump)
-      printf("<pre>\n");
-    if (load_bookmark) {
+
+  if (!isatty(1) && !w3m_dump && !w3m_halfdump) {	/* redirected output */
+    w3m_dump = T;
+    if (COLS == 0)
+      COLS = 80;
+  }
+  if (!w3m_dump && !w3m_halfdump)
+    fmInit();
+  if (w3m_halfdump)
+    printf("<pre>\n");
+
+  if (load_argc == 0) {
+    if (!isatty(0)) {
+      redin = fdopen(dup(0), "rb");
+      newbuf = openGeneralPagerBuffer(redin);
+      dup2(1, 0);
+    } else if (load_bookmark) {
       newbuf = loadGeneralFile(BookmarkFile, NULL,NO_REFERER,0,NULL);
-      if (newbuf == NULL) {
-	fprintf(stderr, "w3m: Can't load bookmark\n");
-      } else {
-	Firstbuf = Currentbuf = newbuf;
-      }
+      if (newbuf == NULL)
+        fprintf(stderr, "w3m: Can't load bookmark\n");
+    } else if ((p = getenv("HTTP_HOME")) != NULL ||
+	(p = getenv("WWW_HOME")) != NULL) {
+      newbuf = loadGeneralFile(p,NULL,NO_REFERER,0,NULL);
+      if (newbuf == NULL)
+        fprintf(stderr, "w3m: Can't load %s\n", p);
+    } else if (visual_start) {
+      Str s_page;
+      s_page = Strnew_charp("<title>W3M startup page</title><center><b>Welcome to ");
+#ifdef JP_CHARSET
+      Strcat_charp(s_page,"<a href='http://ei5nazha.yz.yamagata-u.ac.jp/~aito/w3m/'>");
+#else
+      Strcat_charp(s_page,"<a href='http://ei5nazha.yz.yamagata-u.ac.jp/~aito/w3m/eng'>");
+#endif /* JP_CHARSET */
+      Strcat_m_charp(s_page,
+		   "w3m</a>!<p><p>This is w3m version ",
+		   version,
+		   "<br>Written by <a href='mailto:aito@ei5sun.yz.yamagata-u.ac.jp'>Akinori Ito</a>",
+		   NULL);
+#ifdef DEBIAN
+      Strcat_m_charp(s_page,
+		   "<p>Debian package is maintained by <a href='mailto:ukai@debian.or.jp'>Fumitoshi UKAI</a>.",
+		   "You can read <a href='file:///usr/share/doc/w3m/'>w3m documents on your local system</a>.",
+		   NULL);
+#endif /* DEBIAN */
+      newbuf = loadHTMLString(s_page->ptr);
+      if (newbuf == NULL)
+        fprintf(stderr, "w3m: Can't load string\n");
+      else
+        newbuf->bufferprop |= (BP_INTERNAL | BP_NO_URL);
+    } else {
+      if (fmInitialized)
+	fmTerm();
+      usage();
     }
-    for (i = 0; i < load_argc; i++) {
+    if (newbuf == NULL) {
+      deleteFiles();
+      if (fmInitialized)
+	fmTerm();
+      exit(2);
+    }
+    i = -1;
+  } else {
+    i = 0;
+  }
+  for (; i < load_argc; i++) {
+    if (i >= 0) {
       if (w3m_dump && w3m_dump_head) {
  	request = New(FormList);
  	request->method = FORM_METHOD_HEAD;
@@ -371,114 +413,45 @@ MAIN(int argc, char **argv, char **envp)
       case SCM_MAILTO:
 	break;
       case SCM_LOCAL:
+      case SCM_LOCAL_CGI:
 	insertHist(LoadHist, load_argv[i]);
 	break;
       default:
-	insertHist(URLHist, load_argv[i]);
+	addUniqHist(URLHist, load_argv[i]);
 	break;
       }
-      if (w3m_halfdump) {
-	Currentbuf = Firstbuf = newbuf;
-        printf("</pre><title>%s</title>\n",newbuf->buffername);
-	deleteFiles();
-	exit(0);
-      }
-      if (w3m_dump) {
-	Currentbuf = Firstbuf = newbuf;
-	if (w3m_dump_source) {
-	  dump_source(Currentbuf);
-	}
-	else if (w3m_dump_head) {
-	  dump_head(Currentbuf);
-	}
-	else {
-	  if (Currentbuf->frameset != NULL && RenderFrame) {
-	    rFrame();
-	  }
-	  saveBuffer(Currentbuf,stdout);
-	}
-	deleteFiles();
-#ifdef USE_COOKIE
-        save_cookies();
-#endif
-        exit(0);
-      }
-      if (Currentbuf == NULL)
-	Firstbuf = Currentbuf = newbuf;
-      else {
-	Currentbuf->nextBuffer = newbuf;
-	Currentbuf = newbuf;
-      }
-      if (Currentbuf->frameset != NULL && RenderFrame) {
-	rFrame();
-	Currentbuf = newbuf;
-      }
     }
-  }
-  else {
-    /* When no arguments other than options are given... */
-    if (!isatty(1) || w3m_dump) {		/* redirected output */
-      if (isatty(0)) {
-	fprintf(stderr, "w3m: Can't redirect\n");
-	exit(1);
-      }
-      /* when -T option is given, behave as an HTML formatter */
-      newbuf = openGeneralPagerBuffer(stdin);
-      if (newbuf != NO_BUFFER) {
-	saveBuffer(newbuf,stdout);
-	discardBuffer(newbuf);
-      }
+    if (w3m_halfdump) {
+      Currentbuf = Firstbuf = newbuf;
+      printf("</pre><title>%s</title>\n",newbuf->buffername);
       deleteFiles();
       exit(0);
     }
-    if (!isatty(0)) {		/* redirected input */
-      fmInit();
-      redin = fdopen(dup(0), "r");
-      newbuf = openGeneralPagerBuffer(redin);
-      dup2(1, 0);
-      Firstbuf = Currentbuf = newbuf;
-    }
-  }
-
-  if ((Firstbuf == NULL) && (newbuf == NULL) &&
-      (((p = getenv("HTTP_HOME")) != NULL) ||
-       ((p = getenv("WWW_HOME")) != NULL))) {
-      fmInit();
-      newbuf = loadGeneralFile(p,NULL,NO_REFERER,0,NULL);
-      if (newbuf == NULL)
-	fprintf(stderr, "w3m: Can't load %s\n", p);
+    if (w3m_dump) {
+      Currentbuf = Firstbuf = newbuf;
+      if (w3m_dump_source)
+       dump_source(Currentbuf);
+      else if (w3m_dump_head)
+       dump_head(Currentbuf);
       else {
-	Firstbuf = Currentbuf = newbuf;
-        if (Currentbuf->frameset != NULL && RenderFrame)
+	if (Currentbuf->frameset != NULL && RenderFrame)
 	  rFrame();
+	saveBuffer(Currentbuf,stdout);
       }
-  }
-  if ((Firstbuf == NULL) && (newbuf == NULL) && visual_start) {
-    Str s_page;
-    fmInit();
-    s_page = Strnew_charp("<title>W3M startup page</title><center><b>Welcome to ");
-#ifdef JP_CHARSET
-    Strcat_charp(s_page,"<a href='http://ei5nazha.yz.yamagata-u.ac.jp/~aito/w3m/'>");
-#else
-    Strcat_charp(s_page,"<a href='http://ei5nazha.yz.yamagata-u.ac.jp/~aito/w3m/eng'>");
-#endif
-    Strcat_m_charp(s_page,
-		   "w3m</a>!<p><p>This is w3m version ",
-		   version,
-		   "<br>Written by <a href='mailto:aito@ei5sun.yz.yamagata-u.ac.jp'>Akinori Ito</a>",
-		   NULL);
-#ifdef DEBIAN
-    Strcat_m_charp(s_page,
-		   "<p>Debian package is maintained by <a href='mailto:ukai@debian.or.jp'>Fumitoshi UKAI</a>.",
-		   "You can read <a href='file:///usr/share/doc/w3m/'>w3m documents on your local system</a>.",
-		   NULL);
-#endif
-    newbuf = loadHTMLString(s_page->ptr);
-    if (newbuf == NULL)
-      fprintf(stderr, "w3m: Can't load string\n");
+#ifdef USE_COOKIES
+      save_cookies();
+#endif /* USE_COOKIES */
+      exit(0);
+    }
+    if (Currentbuf == NULL)
+      Firstbuf = Currentbuf = newbuf;
     else {
-      newbuf->bufferprop |= (BP_INTERNAL | BP_NO_URL);
-      Firstbuf = Currentbuf = newbuf;	
+      Currentbuf->nextBuffer = newbuf;
+      Currentbuf = newbuf;
+    }
+    if (Currentbuf->frameset != NULL && RenderFrame) {
+      rFrame();
+      Currentbuf = newbuf;
     }
   }
   if (Firstbuf == NULL) {
@@ -488,12 +461,17 @@ MAIN(int argc, char **argv, char **envp)
       quitfm();
     }
     deleteFiles();
-    fmTerm();
-    usage();
+    if (fmInitialized)
+      fmTerm();
+    exit(2);
   }
+
 #ifdef SIGWINCH
   signal(SIGWINCH, resize_hook);
-#endif
+#else /* not SIGWINCH */
+  setlinescols();
+  setupscreen();
+#endif /* not SIGWINCH */
   Currentbuf = Firstbuf;
   displayBuffer(Currentbuf, B_NORMAL);
   if (line_str){
@@ -504,7 +482,7 @@ MAIN(int argc, char **argv, char **envp)
     if (n_event_queue > 0) {
       for (i = 0; i < n_event_queue; i++) {
 	cmd_argument = eventQueue[i].user_data;
-	eventQueue[i].cmd();
+	w3mFuncList[eventQueue[i].cmd].func();
       }
       n_event_queue = 0;
     }
@@ -512,11 +490,25 @@ MAIN(int argc, char **argv, char **envp)
     /* get keypress event */
 #ifdef MOUSE
     if (use_mouse) mouse_active();
-#endif
+#endif /* MOUSE */
     c = getch();
+#ifdef __EMX__
+    if(!c)
+      switch(getch()){
+	case 73: c=	'b';  break;	// page up
+	case 81: c=	' ';  break;	// page down
+	case 72: c=	'k';  break;	// up arrow
+	case 75: c=	'h';  break;	// left arrow
+	case 77: c=	'l';  break;	// right arrow
+	case 80: c=	'j';  break;	// down arrow
+	case 15: c=ctrl('u'); break;	// shift + tab
+	case 71: c=	'g';  break;	// home
+	case 79: c=	'G';		// end
+      }
+#endif /* __EMX__ */
 #ifdef MOUSE
     if (use_mouse) mouse_inactive();
-#endif
+#endif /* MOUSE */
     if ((c & 0x80) == 0) {	/* Ascii */
       if (('0' <= c) && (c <= '9')) {
 	prec_num = prec_num * 10 + (int)(c - '0');
@@ -533,12 +525,12 @@ MAIN(int argc, char **argv, char **envp)
 static void
 keyPressEventProc(int c)
 {
-    (*GlobalKeymap[c]) ();
+    w3mFuncList[GlobalKeymap[c]].func ();
     onA();
 }
 
 void
-pushEvent(void (*event)(),void *user_data)
+pushEvent(int event,void *user_data)
 {
   if (n_event_queue < N_EVENT_QUEUE) {
     eventQueue[n_event_queue].cmd = event;
@@ -585,7 +577,7 @@ nulcmd(void)
 void
 escmap(void)
 {
-  EscKeymap[(int)getch()] ();
+  w3mFuncList[EscKeymap[(int)getch()]].func ();
 }
 
 void
@@ -597,7 +589,7 @@ escbmap(void)
   if (IS_DIGIT(c))
     escdmap(c);
   else
-    EscBKeymap[(int)c] ();
+    w3mFuncList[EscBKeymap[(int)c]].func ();
 }
 
 void
@@ -612,7 +604,7 @@ escdmap(char c)
     c = getch();
   }
   if (c == '~')
-    EscDKeymap[d] ();
+    w3mFuncList[EscDKeymap[d]].func ();
 }
 
 static void
@@ -668,7 +660,7 @@ resize_hook(SIGNAL_ARG)
   signal(SIGWINCH,resize_hook);
   SIGNAL_RETURN;
 }
-#endif
+#endif /* SIGWINCH */
 
 static void
 curHome(void)
@@ -691,9 +683,9 @@ pgFore(void)
     return;
 #ifdef VI_PREC_NUM
   Currentbuf->topLine = lineSkip(Currentbuf->topLine, PREC_NUM * (LASTLINE - 1));
-#else
+#else /* not VI_PREC_NUM */
   Currentbuf->topLine = lineSkip(Currentbuf->topLine, prec_num ? prec_num : LASTLINE - 1);
-#endif
+#endif /* not VI_PREC_NUM */
   curHome();
 }
 
@@ -705,9 +697,9 @@ pgBack(void)
     return;
 #ifdef VI_PREC_NUM
   Currentbuf->topLine = lineSkip(Currentbuf->topLine, PREC_NUM * (-LASTLINE + 1));
-#else
+#else /* not VI_PREC_NUM */
   Currentbuf->topLine = lineSkip(Currentbuf->topLine, prec_num ? -prec_num : -LASTLINE + 1);
-#endif
+#endif /* not VI_PREC_NUM */
   curHome();
 }
 
@@ -954,7 +946,7 @@ col1R(void)
   if (CharType(p[i]) == PC_KANJI1)
     delta = 2;
   else
-#endif
+#endif /* JP_CHARSET */
     delta = 1;
   if (buf->currentColumn == buf->pos)
     buf->pos += delta;
@@ -981,7 +973,7 @@ col1L(void)
   if (CharType(p[i - 1]) == PC_KANJI2)
     delta = 2;
   else
-#endif
+#endif /* JP_CHARSET */
     delta = 1;
   if ( buf->visualpos == COLS-1 )
     buf->pos = MAX(0, buf->pos-delta);
@@ -1041,50 +1033,21 @@ readsh(void)
   displayBuffer(Currentbuf, B_FORCE_REDRAW);
 }
 
-static void
-cmd_execcmd(char *cmd)
-{
-  int l = strlen(cmd)-1;
-  int bg = 0;
-  fmTerm();
-  printf("\n");
-  while (l > 0) {
-    switch (cmd[l]) {
-    case ' ':
-    case '\t':
-    case '\n':
-      l--;
-      continue;
-    case '&':
-      bg = 1;
-      cmd[l] = '\0';
-      break;
-    default:
-      break;
-    }
-    break;
-  }
-  mySystem(cmd,bg);
-  if (!bg) {
-    printf("[Hit any key]");
-    fflush(stdout);
-    fmInit();
-    getch();
-  }
-  else
-    fmInit();
-  displayBuffer(Currentbuf, B_FORCE_REDRAW);
-}
-
 /* Execute shell command */
 void
 execsh(void)
 {
   char           *cmd;
   cmd = inputLineHist("(exec shell)!", "", IN_COMMAND, ShellHist);
-  if (cmd != NULL && *cmd != '\0')
+  if (cmd != NULL && *cmd != '\0') {
+    fmTerm();
     system(cmd);
-  displayBuffer(Currentbuf, B_NORMAL);
+    printf("[Hit any key]");
+    fflush(stdout);
+    fmInit();
+    getch();
+  }
+  displayBuffer(Currentbuf, B_FORCE_REDRAW);
 }
 
 /* Load file */
@@ -1330,11 +1293,11 @@ quitfm(void)
   deleteFiles();
 #ifdef USE_COOKIE
   save_cookies();
-#endif
+#endif /* USE_COOKIE */
 #ifdef USE_HISTORY
   if (URLSaveHist)
     saveHistory(URLHist);
-#endif
+#endif /* USE_HISTORY */
   exit(0);
 }
 
@@ -1383,7 +1346,7 @@ susp(void)
 {
 #ifndef SIGSTOP
   char *shell;
-#endif
+#endif /* not SIGSTOP */
   move(LASTLINE, 0);
   clrtoeolx();
   refresh();
@@ -1393,9 +1356,9 @@ susp(void)
   if (shell == NULL)
     shell = "/bin/sh";
   system(shell);
-#else
+#else /* SIGSTOP */
   kill(getpid(), SIGSTOP);
-#endif
+#endif /* SIGSTOP */
   fmInit();
   displayBuffer(Currentbuf, B_FORCE_REDRAW);
 }
@@ -1490,7 +1453,7 @@ editBf(void)
   if (strcasestr(Editor,"vi")) {
     Strcat(cmd,Sprintf(" +%d",lnum));
   }
-  Strcat_m_charp(cmd," ",Currentbuf->filename,NULL);
+  Strcat_m_charp(cmd," ",quoteShell(fn)->ptr,NULL);
   fmTerm();
   system(cmd->ptr);
   fmInit();
@@ -1673,7 +1636,7 @@ reMark(void)
 
   displayBuffer(Currentbuf, B_FORCE_REDRAW);
 }
-#endif
+#endif /* USE_MARK */
 
 #ifdef JP_CHARSET
 static char*
@@ -1691,39 +1654,33 @@ cURLcode(char *url, Buffer *buf)
   }
   return url;
 }
-#else
+#else /* not JP_CHARSET */
 #define cURLcode(url,buf) (url)
-#endif
+#endif /* not JP_CHARSET */
+
+static Buffer*
+loadNormalBuf(Buffer *buf, int renderframe)
+{
+  pushBuffer(buf);
+  addUniqHist(URLHist, parsedURL2Str(&Currentbuf->currentURL)->ptr);
+  if (renderframe && RenderFrame && Currentbuf->frameset != NULL)
+    rFrame();
+  return buf;
+}
 
 static Buffer *
 loadLink(char *url, char *target, char *referer, FormList *request)
 {
-  Buffer *buf;
-  struct frame_body *f_body;
-  int flag;
+  Buffer *buf, *nfbuf;
+  union frameset_element *f_element = NULL;
+  int flag = 0;
 
   message(Sprintf("loading %s\n", url)->ptr, 0, 0);
   refresh();
 
-  if (Currentbuf->bufferprop & BP_FRAME && target != NULL && !do_download) {
-    buf = Currentbuf->linkBuffer[LB_N_FRAME];
-    if (buf == NULL)
-      goto normal_load;
-    f_body = search_frame(buf->frameset,target);
-    if (f_body == NULL)
-      goto normal_load;
-    delBuffer(Currentbuf);
-    Currentbuf = buf;
-    if (f_body->to_delete)
-      unlink(f_body->source);
-    f_body->source = NULL;
-    f_body->url = url;
-    f_body->referer = referer;
-    rFrame();
-    return buf;
+  if (referer == NULL) {
+    referer = parsedURL2Str(&Currentbuf->currentURL)->ptr;
   }
-
-normal_load:
   buf = loadGeneralFile(cURLcode(url,Currentbuf),
 			baseURL(Currentbuf),
 			referer,
@@ -1731,12 +1688,46 @@ normal_load:
 			request);
   if (buf == NULL) {
     disp_message(Sprintf("Can't load %s\n", url)->ptr, NIL);
-  } else if (buf != NO_BUFFER) {
-    pushBuffer(buf);
-    addUniqHist(URLHist, parsedURL2Str(&Currentbuf->currentURL)->ptr);
-    if (RenderFrame && Currentbuf->frameset != NULL)
-      rFrame();
+    return NULL;
+  } 
+  if (buf == NO_BUFFER) {
+    return NULL;
   }
+  if (!on_target) /* open link as an indivisual page */
+    return loadNormalBuf(buf,T);
+
+  if (do_download) /* download (thus no need to render frame) */
+    return loadNormalBuf(buf,NIL);
+
+  if (target == NULL || /* no target specified
+			   (that means this page is not a frame page) */
+      !strcmp(target, "_top") || /* this link is specified to be opened
+				    as an indivisual page */
+      !(Currentbuf->bufferprop & BP_FRAME) /* This page is not a frame page */
+      ) {
+    return loadNormalBuf(buf,T);
+  }
+  nfbuf = Currentbuf->linkBuffer[LB_N_FRAME];
+  if (nfbuf == NULL) {
+    /* original page (that contains <frameset> tag) doesn't exist */
+    return loadNormalBuf(buf,T);
+  }
+  
+  f_element = search_frame(nfbuf->frameset, target);
+  if (f_element == NULL) {
+    /* specified target doesn't exist in this frameset */
+    return loadNormalBuf(buf,T);
+  }
+
+  /* frame page */
+  delBuffer(Currentbuf);
+  Currentbuf = nfbuf;
+  /* save current frame combination */
+  pushFrameTree(&(nfbuf->frameQ), copyFrameSet(nfbuf->frameset));
+  /*  nfbuf->frameset = copyFrameSet(nfbuf->frameset); */
+  resetFrameElement(f_element,buf,referer,request);
+  discardBuffer(buf);
+  rFrame();
   displayBuffer(Currentbuf, B_NORMAL);
   return buf;
 }
@@ -1749,6 +1740,8 @@ followA(void)
   Anchor         *a;
   Buffer         *buf;
   int             ln,i;
+  char           *local_label = NULL;
+  ParsedURL       u;
 
   if (Currentbuf->firstLine == NULL)
     return;
@@ -1760,7 +1753,19 @@ followA(void)
     return;
   }
   if (*a->url == '#') {		/* index within this buffer */
-    ln = searchURLLabelLine(Currentbuf, a->url + 1);
+    local_label = a->url+1;
+  }
+  parseURL2(a->url,&u,baseURL(Currentbuf));
+  if (u.scheme == Currentbuf->currentURL.scheme &&
+      u.port == Currentbuf->currentURL.port &&
+      strcasecmp(u.host,Currentbuf->currentURL.host) == 0 &&
+      strcmp(u.file,Currentbuf->currentURL.file) == 0) {
+    /* index within this buffer */
+    local_label = u.label;
+  }
+
+  if (local_label != NULL) {
+    ln = searchURLLabelLine(Currentbuf, local_label);
     if (ln < 0) {
       disp_message("Not found",T);
       return;
@@ -1780,19 +1785,33 @@ followA(void)
   if (!strncasecmp(a->url,"mailto:",7)) {
     /* invoke external mailer */
     Str tmp;
-    tmp = Strnew_m_charp(Mailer," ",a->url+7,NULL);
-    cmd_execcmd(tmp->ptr);
+    tmp = Strnew_m_charp(Mailer," ",quoteShell(a->url+7)->ptr,NULL);
+    fmTerm();
+    system(tmp->ptr);
+    fmInit();
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
     return;
   }
+#ifdef USE_NNTP
   else if (!strncasecmp(a->url,"news:",5) &&
 	   strchr(a->url,'@') == NULL) {
     /* news:newsgroup is not supported */
     disp_message("news:newsgroup_name is not supported",T);
     return;
   }
+#endif /* USE_NNTP */
   loadLink(a->url,a->target,a->referer,NULL);
 }
 
+/* follow HREF link in the buffer */
+void
+bufferA(void)
+{
+  on_target = NIL;
+  followA();
+  on_target = T;
+}
+
 /* view inline image */
 void
 followI(void)
@@ -1838,7 +1857,7 @@ save_submit_formlist(FormItemList *src)
   FormSelectOptionItem *opt;
   FormSelectOptionItem *curopt;
   FormSelectOptionItem *srcopt;
-#endif
+#endif /* MENU_SELECT */
 
   if (src == NULL)
     return NULL;
@@ -1882,7 +1901,7 @@ save_submit_formlist(FormItemList *src)
     item->select_option = opt;
     if (srcitem->label)
       item->label = Strdup(srcitem->label);
-#endif
+#endif /* MENU_SELECT */
     item->parent = list;
     item->next = NULL;
 
@@ -1904,7 +1923,7 @@ static void
 query_from_followform(Str *query, FormItemList *fi, int multipart)
 {
   FormItemList *f2;
-  FILE *body;
+  FILE *body = NULL;
 
   if (multipart) {
     *query = Sprintf("%s/w3mform%d.%lx", rc_dir, getpid(),
@@ -1949,9 +1968,9 @@ query_from_followform(Str *query, FormItemList *fi, int multipart)
 	    *query = conv(f2->value->ptr,InnerCode,fi->parent->charset);
 	  else
 	    *query = conv(f2->value->ptr,InnerCode,Currentbuf->document_code);
-  #else
+  #else /* not JP_CHARSET */
 	  *query = f2->value;
-  #endif
+  #endif /* not JP_CHARSET */
 	  if (f2->type == FORM_INPUT_FILE)
 	    form_write_form_file(body, fi->parent->boundary, f2->name->ptr, (*query)->ptr);
 	  else
@@ -1976,9 +1995,9 @@ query_from_followform(Str *query, FormItemList *fi, int multipart)
 	      Strcat(*query,form_quote(conv(f2->value->ptr,InnerCode,fi->parent->charset)));
 	    else
 	      Strcat(*query,form_quote(conv(f2->value->ptr,InnerCode,Currentbuf->document_code)));
-  #else
+  #else /* not JP_CHARSET */
 	    Strcat(*query,form_quote(f2->value));
-  #endif
+  #endif /* not JP_CHARSET */
 	  }
 	}
 	if (f2->next)
@@ -2064,7 +2083,7 @@ followForm(void)
       Currentbuf->cursorY);
     formUpdateBuffer(a,Currentbuf,fi);
     break;
-#endif
+#endif /* MENU_SELECT */
   case FORM_INPUT_IMAGE:
   case FORM_INPUT_SUBMIT:
   case FORM_INPUT_BUTTON:
@@ -2082,7 +2101,7 @@ followForm(void)
     if (fi->parent->method == FORM_METHOD_GET) {
       Strcat_charp(tmp2,"?");
       Strcat(tmp2,tmp);
-      loadLink(tmp2->ptr,fi->parent->target,NULL,NULL);
+      loadLink(tmp2->ptr,a->target,NULL,NULL);
     }
     else if (fi->parent->method == FORM_METHOD_POST) {
       Buffer *buf;
@@ -2094,7 +2113,7 @@ followForm(void)
 	fi->parent->body = tmp->ptr;
 	fi->parent->length = tmp->length;
       }
-      buf = loadLink(tmp2->ptr,fi->parent->target,NULL,fi->parent);
+      buf = loadLink(tmp2->ptr,a->target,NULL,fi->parent);
       if (multipart) {
 	unlink(fi->parent->body);
       }
@@ -2381,8 +2400,11 @@ checkBackBuffer(Buffer *buf)
 
   if (nbuf == NULL)
     return NIL;
-  else if (RenderFrame && nbuf == buf->linkBuffer[LB_N_FRAME])
+  else if (RenderFrame && nbuf == buf->linkBuffer[LB_N_FRAME]) {
+    if (nbuf->frameQ != NULL)
+      return T;
     return checkBackBuffer(nbuf);
+  }
   else
     return T;
 }
@@ -2394,8 +2416,17 @@ realBackBuffer()
 
   buf = Currentbuf->linkBuffer[LB_N_FRAME];
   delBuffer(Currentbuf);
-  if (RenderFrame && Currentbuf == buf)
-    realBackBuffer();
+  if (RenderFrame && Currentbuf == buf) {
+    struct frameset *fs;
+
+    fs = popFrameTree(&(buf->frameQ));
+    if (fs == NULL)
+      realBackBuffer();
+    else {
+      deleteFrameSet(buf->frameset);
+      buf->frameset = fs;
+    }
+  }
 }
 
 /* delete current buffer and back to the previous buffer */
@@ -2407,6 +2438,8 @@ backBf(void)
     return;
   }
   realBackBuffer();
+  if (RenderFrame && Currentbuf->frameset != NULL)
+    rFrame();
   clear();
   displayBuffer(Currentbuf, B_FORCE_REDRAW);
 }
@@ -2424,14 +2457,16 @@ cmd_loadURL(char *url, ParsedURL *current)
 {
   Buffer         *buf;
 
+#ifdef USE_NNTP
   if (!strncasecmp(url,"news:",5) &&
 	   strchr(url,'@') == NULL) {
     /* news:newsgroup is not supported */
     disp_message("news:newsgroup_name is not supported",T);
     return;
   }
+#endif /* USE_NNTP */
   
-  message(Sprintf("loading %s\n", url)->ptr, 0, 0);
+/*  message(Sprintf("loading %s\n", url)->ptr, 0, 0); */
   refresh();
   buf = loadGeneralFile(url, current,NO_REFERER,0,NULL);
   if (buf == NULL) {
@@ -2453,8 +2488,11 @@ goURL(void)
 
   if (cmd_argument)
     url = (char*)cmd_argument;
-  else
+  else {
+    if (! (Currentbuf->bufferprop & BP_INTERNAL))
+      addUniqHist(URLHist, parsedURL2Str(&Currentbuf->currentURL)->ptr);
     url = inputStrHist("Goto URL: ", NULL, URLHist);
+  }
   if (url == NULL || *url == '\0') {
     displayBuffer(Currentbuf, B_FORCE_REDRAW);
     return;
@@ -2496,8 +2534,13 @@ adBmark(void)
   /*  cmd_loadBuffer(load_bookmark_panel(Currentbuf), BP_NO_URL, LB_NOLINK); */
   url = form_quote(parsedURL2Str(&Currentbuf->currentURL));
   title = form_quote(Strnew_charp(Currentbuf->buffername));
+#ifdef __EMX__
+  tmp = Sprintf("%s/w3mbookmark.exe?mode=panel&bmark=%s&url=%s&title=%s",
+		get_os2_dft("W3M_LIB_DIR",LIB_DIR),BookmarkFile,url->ptr,title->ptr);
+#else /* not __EMX__ */
   tmp = Sprintf("%s/w3mbookmark?mode=panel&bmark=%s&url=%s&title=%s",
 		LIB_DIR,BookmarkFile,url->ptr,title->ptr);
+#endif /* not __EMX__ */
   cmd_loadURL(tmp->ptr,NULL);
 }
 
@@ -2525,7 +2568,7 @@ pginfo(void)
 #ifdef JP_CHARSET
   if (buf != NULL)
     buf->document_code = Currentbuf->document_code;
-#endif
+#endif /* JP_CHARSET */
   cmd_loadBuffer(buf, BP_NORMAL, LB_INFO);
 }
 
@@ -2550,7 +2593,7 @@ cooLst(void)
   if (buf != NULL)
     cmd_loadBuffer(buf, BP_NO_URL, LB_NOLINK);
 }
-#endif
+#endif /* USE_COOKIE */
 
 #ifdef USE_HISTORY
 /* History page */
@@ -2559,7 +2602,7 @@ ldHist(void)
 {
   cmd_loadBuffer(historyBuffer(URLHist), BP_NO_URL, LB_NOLINK);
 }
-#endif
+#endif /* USE_HISTORY */
 
 /* download HREF link */
 void
@@ -2641,11 +2684,17 @@ peekURL(void)
   a = retrieveCurrentAnchor(Currentbuf);
   if (a == NULL) {
     a = retrieveCurrentImg(Currentbuf);
-    if (a == NULL)
-      return;
+    if (a == NULL) {
+      a = retrieveCurrentForm(Currentbuf);
+      if (a == NULL)
+        return;
+      s = Strnew_charp(form2str((FormItemList*)a->url));
+      goto disp;
+    }
   }
   parseURL2(a->url,&pu,baseURL(Currentbuf));
   s = parsedURL2Str(&pu);
+disp:
   disp_message(s->ptr,T);
 }
 
@@ -2732,7 +2781,7 @@ reload(void)
 {
   Buffer *buf, *fbuf = NULL;
   Str url;
-  int linenum, is_frame, is_source;
+  int linenum, is_source;
   FormList *request;
   int multipart;
 
@@ -2746,17 +2795,31 @@ reload(void)
     disp_message("Can't reload stdin",T);
     return;
   }
-  is_frame = Currentbuf->bufferprop & BP_FRAME;
-  if (is_frame)
-    fbuf = Currentbuf->linkBuffer[LB_N_FRAME];
-  else if (Currentbuf->frameset != NULL)
-    fbuf = Currentbuf->linkBuffer[LB_FRAME];
-  is_source = Currentbuf->bufferprop & BP_SOURCE;
   if (Currentbuf->firstLine == NULL)
     linenum = 1;
   else
     linenum = Currentbuf->currentLine->linenumber;
+  if (Currentbuf->bufferprop & BP_FRAME &&
+      (fbuf = Currentbuf->linkBuffer[LB_N_FRAME])) {
+    if (fmInitialized) {
+      message("Rendering frame",0,0);
+      refresh();
+    }
+    if (!(buf = renderFrame(fbuf)))
+      return;
+    delBuffer(fbuf->linkBuffer[LB_FRAME]);
+    fbuf->linkBuffer[LB_FRAME] = buf;
+    buf->linkBuffer[LB_N_FRAME] = fbuf;
+    pushBuffer(buf);
+    Currentbuf = buf;
+    gotoLine(Currentbuf,linenum);
+    arrangeCursor(Currentbuf);
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+    return;
+  } else if (Currentbuf->frameset != NULL)
+    fbuf = Currentbuf->linkBuffer[LB_FRAME];
   multipart = 0;
+  is_source = Currentbuf->bufferprop & BP_SOURCE;
   if (Currentbuf->form_submit) {
     request = Currentbuf->form_submit->parent;
     if (request->method == FORM_METHOD_POST
@@ -2790,8 +2853,6 @@ reload(void)
     vwSrc();
     if (Currentbuf != buf)
       Firstbuf = deleteBuffer(Firstbuf, buf);
-  } else if (is_frame && buf->frameset != NULL) {
-    rFrame();
   }
   gotoLine(Currentbuf,linenum);
   arrangeCursor(Currentbuf);
@@ -2806,10 +2867,12 @@ chkURL(void)
     "http://[a-zA-Z0-9][a-zA-Z0-9:%\\-\\./?=~_\\&+@#,]*",
 #ifdef USE_SSL
     "https://[a-zA-Z0-9][a-zA-Z0-9:%\\-\\./?=~_\\&+@#,]*",
-#endif
+#endif /* USE_SSL */
     "gopher://[a-zA-Z0-9][a-zA-Z0-9:%\\-\\./_]*",
     "ftp://[a-zA-Z0-9][a-zA-Z0-9%\\-\\./_+:@]*",
+#ifdef USE_NNTP
     "news:[^<> 	][^<> 	]*",
+#endif /* not USE_NNTP */
     NULL,
   };
   int             i;
@@ -2820,6 +2883,7 @@ chkURL(void)
   displayBuffer(Currentbuf, B_FORCE_REDRAW);
 }
 
+#ifdef USE_NNTP
 /* mark Message-ID-like patterns as NEWS anchors */
 void
 chkNMID(void)
@@ -2835,6 +2899,7 @@ chkNMID(void)
 
   displayBuffer(Currentbuf, B_FORCE_REDRAW);
 }
+#endif /* USE_NNTP */
 
 /* render frame */
 void
@@ -2888,21 +2953,23 @@ invoke_browser(char *url)
     break;
   }
   if (browser == NULL || *browser == '\0') {
-    tmp = Strnew_charp(inputStr("Browse cmd:",url));
+    browser = inputStr("Browse command: ", NULL);
+    if (browser == NULL || *browser == '\0')
+      return;
   }
-  else if (strcasestr(browser,"%s")) {
+  url = quoteShell(url)->ptr;
+  if (strcasestr(browser,"%s")) {
     tmp = Sprintf(browser,url);
-  }
-  else {
+    Strremovetrailingspaces(tmp);
+    if (Strlastchar(tmp) == '&') {
+      Strshrink(tmp,1);
+      bg = 1;
+    }
+  } else {
     tmp = Strnew_charp(browser);
     Strcat_char(tmp,' ');
     Strcat_charp(tmp,url);
   }
-  Strremovetrailingspaces(tmp);
-  if (Strlastchar(tmp) == '&') {
-    Strshrink(tmp,1);
-    bg = 1;
-  }
   fmTerm();
   mySystem(tmp->ptr,bg);
   fmInit();
@@ -2953,9 +3020,9 @@ void curlno()
 #ifdef JP_CHARSET
   tmp = Sprintf("line %d/%d (%d%%) code %c",cur,all,cur*100/all,
                 Currentbuf->document_code);
-#else
+#else /* not JP_CHARSET */
   tmp = Sprintf("line %d/%d (%d%%)",cur,all,cur*100/all);
-#endif
+#endif /* not JP_CHARSET */
                 
   disp_message(tmp->ptr,NIL);
 }
@@ -2998,12 +3065,12 @@ process_mouse(int btn, int x, int y)
 
 	  if (delta_x > 0) {
 	    for (i = 0; i < delta_x; i++)
-              if(reverse_mouse) col1L();
-              else col1R();
-	  } else if (delta_x < 0) {
-	    for (i = 0; i < -delta_x; i++)
               if(reverse_mouse) col1R();
               else col1L();
+	  } else if (delta_x < 0) {
+	    for (i = 0; i < -delta_x; i++)
+              if(reverse_mouse) col1L();
+              else col1R();
 	  }
 	} else {
 	  if (y == LASTLINE) {
@@ -3025,7 +3092,7 @@ process_mouse(int btn, int x, int y)
 #ifdef JP_CHARSET
 	       || ((Currentbuf->currentLine->propBuf[Currentbuf->pos] & PC_KANJI1)
 		    && x == Currentbuf->cursorX + 1)
-#endif
+#endif /* JP_CHARSET */
 	       )) {
 	    followA();
 	    return;
@@ -3044,7 +3111,7 @@ process_mouse(int btn, int x, int y)
 	cursorXY(Currentbuf, x, y);
 	onA();
 	mainMenu(x, y);
-#endif
+#endif /* MENU */
 	break;
     }
   } else {
@@ -3103,8 +3170,27 @@ int gpm_process_mouse(Gpm_Event *event, void *data)
   process_mouse(btn, x-1, y-1);
   return 0;
 }
-#endif
-#endif
+#endif /* USE_GPM */
+
+#ifdef USE_SYSMOUSE
+int sysm_process_mouse(int x, int y, int nbs, int obs)
+{
+  int btn;
+  int bits;
+
+  if (obs & ~nbs)
+    btn = MOUSE_BTN_UP;
+  else if (bits = nbs & ~obs)
+    btn = bits & 0x1 ? MOUSE_BTN1_DOWN :
+         (bits & 0x2 ? MOUSE_BTN2_DOWN :
+         (bits & 0x4 ? MOUSE_BTN3_DOWN : 0));
+  else /* nbs == obs */
+    return 0;
+  process_mouse(btn, x, y);
+  return 0;
+}
+#endif /* USE_SYSMOUSE */
+#endif /* MOUSE */
 
 void wrapToggle( void )
 {
@@ -3182,4 +3268,4 @@ dictwordat(void)
 {
   execdict(GetWord(Currentbuf));
 }
-#endif
+#endif /* DICT */
