@@ -12,11 +12,18 @@
 #include <gpm.h>
 static int gpm_process_menu_mouse(Gpm_Event *event, void *data);
 extern int gpm_process_mouse(Gpm_Event*,void*);
-#define X_GPM_SELECTED (char)0xff
-static int X_Gpm_Selection;
+#endif /* USE_GPM */
+#ifdef USE_SYSMOUSE
+extern int (*sysm_handler)(int x, int y, int nbs, int obs);
+static int sysm_process_menu_mouse(int, int, int, int);
+extern int sysm_process_mouse(int, int, int, int);
+#endif /* USE_SYSMOUSE */
+#if defined(USE_GPM) || defined(USE_SYSMOUSE)
+#define X_MOUSE_SELECTED (char)0xff
+static int X_Mouse_Selection;
 extern int do_getch();
 #define getch()	do_getch()
-#endif
+#endif /* defined(USE_GPM) || defined(USE_SYSMOUSE) */
 
 #ifdef MENU
 
@@ -26,18 +33,18 @@ static char *FRAME[] = {
   "", "", "", 
   "", "  ", "", 
   "", "", "",
-#else
+#else /* not MENU_THIN_FRAME */
   "", "", "", 
   "", "  ", "", 
   "", "", "",
-#endif
+#endif /* not MENU_THIN_FRAME */
   "", ""};
 #define FRAME_WIDTH 2
 
 #define G_start  /**/
 #define G_end    /**/
 
-#else
+#else /* not KANJI_SYMBOLS */
 static char *N_FRAME[] = {
   "+", "-", "+",
   "|", " ", "|",
@@ -56,7 +63,7 @@ static int graph_mode = NIL;
 
 #define G_start  {if (graph_mode) graphstart();}
 #define G_end    {if (graph_mode) graphend();}
-#endif
+#endif /* not KANJI_SYMBOLS */
 
 static int mEsc(char c);
 static int mEscB(char c);
@@ -69,6 +76,8 @@ static int mLast(char c);
 static int mTop(char c);
 static int mNext(char c);
 static int mPrev(char c);
+static int mFore(char c);
+static int mBack(char c);
 static int mOk(char c);
 static int mCancel(char c);
 static int mClose(char c);
@@ -157,7 +166,7 @@ static int (*MenuEscBKeymap[128])(char c) = {
 };
 static int (*MenuEscDKeymap[128])(char c) = {
 /* 0      1       INS     3       4       PgUp,   PgDn    7     */
-  mNull,  mNull,  mClose, mNull,  mNull,  mUp,    mDown,  mNull, 
+  mNull,  mNull,  mClose, mNull,  mNull,  mBack,  mFore,  mNull, 
 /* 8      9       10      F1      F2      F3      F4      F5       */
   mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull, 
 /* 16     F6      F7      F8      F9      F10     22      23       */
@@ -216,7 +225,7 @@ static MenuItem MainMenuItem[] = {
   { MENU_FUNC,  "λ         (q)", NULL, 0, qquitfm, NULL, "qQ" },
   { MENU_END,   "",                 NULL, 0, nulcmd,  NULL, ""   },
 };
-#else
+#else /* LANG != JA */
 static MenuItem MainMenuItem[] = {
 /*  type        label           variable value func     popup keys  */
   { MENU_FUNC,  " Back         (b) ", NULL, 0, backBf,  NULL, "b"  },
@@ -239,7 +248,7 @@ static MenuItem MainMenuItem[] = {
   { MENU_FUNC,  " Quit         (q) ", NULL, 0, qquitfm, NULL, "qQ" },
   { MENU_END,   "",                   NULL, 0, nulcmd,  NULL, ""   },
 };
-#endif
+#endif /* LANG != JA  */
 
 /* --- MainMenu (END) --- */
 
@@ -361,7 +370,7 @@ draw_menu(Menu *menu)
       FRAME = N_FRAME;
     }
   }
-#endif
+#endif /* not KANJI_SYMBOLS */
 
   if (menu->offset == 0) {
     G_start;
@@ -512,17 +521,25 @@ action_menu(Menu *menu)
   while(1) {
 #ifdef MOUSE
     if (use_mouse) mouse_active();
-#endif
+#endif /* MOUSE */
     c = getch();
+#ifdef __EMX__
+    if(!c)
+      switch(getch()){
+	case  72: c='k'; break; //down arrow
+	case  80: c='j';	//up arrow
+      }
+#endif /* __EMX__ */
 #ifdef MOUSE
     if (use_mouse) mouse_inactive();
-#endif
-#ifdef USE_GPM
-    if (c == X_GPM_SELECTED) {
-      select = X_Gpm_Selection;
-      break;
+#endif /* MOUSE */
+#if defined(USE_GPM) || defined(USE_SYSMOUSE)
+    if (c == X_MOUSE_SELECTED) {
+      select = X_Mouse_Selection;
+      if (select != MENU_NOTHING)
+        break;
     }
-#endif
+#endif /*  defined(USE_GPM) || defined(USE_SYSMOUSE) */
     if ((c & 0x80) == 0) {      /* Ascii */
       select = (*menu->keymap[(int)c])(c);
       if (select != MENU_NOTHING)
@@ -560,7 +577,10 @@ popup_menu(Menu *parent, Menu *menu)
 
 #ifdef USE_GPM
   gpm_handler = gpm_process_menu_mouse;
-#endif
+#endif /* USE_GPM */
+#ifdef USE_SYSMOUSE
+  sysm_handler = sysm_process_menu_mouse;
+#endif /* USE_SYSMOUSE */
   menu->parent = parent;
   menu->select = menu->initial;
   menu->offset = 0;
@@ -582,7 +602,11 @@ popup_menu(Menu *parent, Menu *menu)
 #ifdef USE_GPM
   if (CurrentMenu == NULL)
     gpm_handler = gpm_process_mouse;
-#endif
+#endif /* USE_GPM */
+#ifdef USE_SYSMOUSE
+  if (CurrentMenu == NULL)
+    sysm_handler = sysm_process_mouse;
+#endif /* USE_SYSMOUSE */
 }
 
 void
@@ -737,6 +761,24 @@ mPrev(char c)
   return(MENU_NOTHING);
 }
 
+static int
+mFore(char c)
+{
+  if (CurrentMenu->select >= CurrentMenu->nitem - 1)
+    return(MENU_NOTHING);
+  goto_menu(CurrentMenu, (CurrentMenu->select + CurrentMenu->height - 1), (CurrentMenu->height + 1));
+  return(MENU_NOTHING);
+}
+
+static int
+mBack(char c)
+{
+  if (CurrentMenu->select <= 0)
+    return(MENU_NOTHING);
+  goto_menu(CurrentMenu, (CurrentMenu->select - CurrentMenu->height + 1), (-1 - CurrentMenu->height));
+  return(MENU_NOTHING);
+}
+
 static int
 mOk(char c)
 {
@@ -854,23 +896,42 @@ gpm_process_menu_mouse(Gpm_Event *event, void *data)
     }
   }
   else {
+    GPM_DRAWPOINTER(event);
     return 0;
   }
   x = event->x;
   y = event->y;
-  X_Gpm_Selection = process_mMouse(btn, x-1, y-1);
-  if (X_Gpm_Selection >= 0)
-    return X_GPM_SELECTED;
-  return 0;
+  X_Mouse_Selection = process_mMouse(btn, x-1, y-1);
+  return X_MOUSE_SELECTED;
 }
-#endif
-#else
+#endif /* USE_GPM */
+
+#ifdef USE_SYSMOUSE
+static int 
+sysm_process_menu_mouse(int x, int y, int nbs, int obs)
+{
+  int btn;
+  int bits;
+
+  if (obs & ~nbs)
+    btn = MOUSE_BTN_UP;
+  else if (bits = nbs & ~obs)
+    btn = bits & 0x1 ? MOUSE_BTN1_DOWN :
+         (bits & 0x2 ? MOUSE_BTN2_DOWN :
+         (bits & 0x4 ? MOUSE_BTN3_DOWN : 0));
+  else /* nbs == obs */
+    return 0;
+  X_Mouse_Selection = process_mMouse(btn, x, y);
+  return X_MOUSE_SELECTED;
+}
+#endif /* USE_GPM */ 
+#else /* not MOUSE */
 static int
 mMouse(char c)
 {
   return(MENU_NOTHING);
 }
-#endif
+#endif /* not MOUSE */
 
 /* --- MenuFunctions (END) --- */
 
@@ -919,6 +980,7 @@ initSelectMenu(void) {
     if (buf->filename != NULL) {
       switch(buf->currentURL.scheme) {
 	case SCM_LOCAL:
+      case SCM_LOCAL_CGI:
 	if (strcmp(buf->currentURL.file, "-")) {
           Strcat_char(str, ' ');
           Strcat_charp(str, buf->filename);
@@ -1056,7 +1118,7 @@ int
 setMenuItem(MenuItem *item, char *type, char *line)
 {
   char *label, *func, *popup, *keys;
-  FuncList *f;
+  int f;
   int n;
 
   if (type == NULL || *type == '\0')	/* error */
@@ -1077,7 +1139,7 @@ setMenuItem(MenuItem *item, char *type, char *line)
     item->type  = MENU_FUNC;
     item->label = label;
     f = getFuncList(func, w3mFuncList, w3mNFuncList);
-    item->func = (f != NULL) ? f->func : nulcmd;
+    item->func = w3mFuncList[(f >= 0) ? f : FUNCNAME_nulcmd].func;
     item->keys  = keys;
     return MENU_FUNC;
   } else if (strcmp(type, "popup") == 0) {
