@@ -49,4 +49,4 @@ typedef struct _MenuList {
 	MenuItem *item;
 } MenuList;
 
-#endif
+#endif /* not MENU_H */
