@@ -10,4 +10,4 @@ extern struct parsed_tagarg * parse_tag ( char *argbuf );
 extern char * tag_get_value ( struct parsed_tagarg *t, char *arg );
 extern int tag_exists ( struct parsed_tagarg *t, char *arg );
 extern struct parsed_tagarg * cgistr2tagarg ( char *cgistr );
-#endif
+#endif /* not PARSETAG_H */
