@@ -8,12 +8,16 @@
 #include "parsetag.h"
 #include "local.h"
 
+#ifdef __EMX__
+#include "os2emx.h"
+#endif /* __EMX__ */
+
 static int rc_initialized = 0;
 char *rc_dir;
 
 #ifdef USE_HISTORY
 extern Hist *URLHist;
-#endif
+#endif /* USE_HISTORY */
 
 #define P_INT      0
 #define P_SHORT    1
@@ -29,6 +33,7 @@ extern Hist *URLHist;
 #define CMT_SAVEHIST     "URL¸"
 #define CMT_KANJICODE    "ɽѴ"
 #define CMT_FRAME        "ե졼μưɽ"
+#define CMT_TSELF        "target̤ξ_selfѤ"
 #define CMT_DISPLINK     "μưɽ"
 #define CMT_MULTICOL     "ե̾Υޥɽ"
 #define CMT_COLOR        "顼ɽ"
@@ -60,6 +65,8 @@ extern Hist *URLHist;
 #define CMT_ACCEPTLANG   "Ĥ(Accept-Language:)"
 #define CMT_WRAP         "ޤ֤"
 #define CMT_USECOOKIE "åѤ"
+#define CMT_ACCEPTCOOKIE "åդ"
+#define CMT_VIEW_UNSEENOBJECTS "طʲؤΥ󥯤"
 #define CMT_BGEXTVIEW    "ӥ塼Хå饦ɤư"
 #define CMT_EXT_DIRLIST  "ǥ쥯ȥꥹȤ˳ޥɤȤ"
 #define CMT_DIRLIST_CMD  "ǥ쥯ȥꥹѥޥ"
@@ -67,9 +74,9 @@ extern Hist *URLHist;
 #ifdef MOUSE
 #define CMT_MOUSE         "ޥȤ"
 #define CMT_REVERSE_MOUSE "ޥΥɥåưդˤ"
-#endif
+#endif /* MOUSE */
 #define CMT_NOSENDREFERER "Referer: ʤ褦ˤ"
-#else
+#else /* LANG != JA */
 #define CMT_HELPER	 "External Viewer Setup"
 #define CMT_TABSTOP      "Tab width"
 #define CMT_PAGERLINE    "# of reserved line when w3m is used as a pager"
@@ -77,6 +84,7 @@ extern Hist *URLHist;
 #define CMT_SAVEHIST     "Save URL history"
 /* #define CMT_KANJICODE    "Display Kanji Code" */
 #define CMT_FRAME        "Automatic rendering of frame"
+#define CMT_TSELF        "use _self as default target"
 #define CMT_DISPLINK     "Automatic display of link URL"
 #define CMT_MULTICOL     "Multi-column output of file names"
 #define CMT_COLOR        "Display with color"
@@ -108,6 +116,8 @@ extern Hist *URLHist;
 #define CMT_ACCEPTLANG   "Accept-Language"
 #define CMT_WRAP         "Wrap search"
 #define CMT_USECOOKIE   "Use Cookie"
+#define CMT_ACCEPTCOOKIE "Accept Cookie"
+#define CMT_VIEW_UNSEENOBJECTS "Display unseenobjects (e.g. bgimage) tag"
 #define CMT_BGEXTVIEW    "Background an external viewer"
 #define CMT_EXT_DIRLIST  "Use external program for directory listing"
 #define CMT_DIRLIST_CMD  "Directory listing command"
@@ -115,9 +125,9 @@ extern Hist *URLHist;
 #ifdef MOUSE
 #define CMT_MOUSE         "Use mouse"
 #define CMT_REVERSE_MOUSE "Reverse mouse dragging action"
-#endif
+#endif /* MOUSE */
 #define CMT_NOSENDREFERER "Don't send header `Referer:'"
-#endif
+#endif /* LANG != JA */
 
 #define PI_TEXT    0
 #define PI_ONOFF   1
@@ -132,17 +142,17 @@ static char *kcodestr[] = {
   "m - JIS X0208-1978 + US_ASCII",
   "n - JIS X0208-1983 + US_ASCII",
   NULL};
-#endif
+#endif /* JP_CHARSET */
 
 #ifdef COLOR
 static char *colorstr[] = {
 #if LANG== JA
   "0:","1:","2:","3:","4:","5:","6:","7:","8:ü",NULL
-#else
+#else /* LANG != JA */
   "0:black","1:red","2:green","3:yellow","4:blue","5:magenta","6:cyan","7:white","8:terminal",NULL
-#endif
+#endif /* LANG != JA */
   };
-#endif
+#endif /* COLOR */ 
 
 #ifdef INET6
 static char *dnsorders[] = {
@@ -151,7 +161,7 @@ static char *dnsorders[] = {
     "2:inet6 inet",
     NULL,
 };
-#endif
+#endif /* INET6 */
 
 struct param_section {
   char *name;
@@ -171,8 +181,9 @@ static struct param_ptr params1[] = {
   {"tabstop",         P_INT,    PI_TEXT,  (void*)&Tabstop,CMT_TABSTOP,NULL},
 #ifdef JP_CHARSET
   {"kanjicode",       P_CHAR,   PI_SEL_C, (void*)&DisplayCode, CMT_KANJICODE,kcodestr},
-#endif
+#endif /* JP_CHARSET */
   {"frame",           P_CHARINT,PI_ONOFF, (void*)&RenderFrame, CMT_FRAME, NULL},
+  {"target_self",    P_CHARINT,PI_ONOFF, (void*)&TargetSelf, CMT_TSELF, NULL},
 {"display_link",    P_INT,    PI_ONOFF, (void*)&displayLink, CMT_DISPLINK,NULL},
   {"ext_dirlist",     P_INT,    PI_ONOFF, (void*)&UseExternalDirBuffer,CMT_EXT_DIRLIST,NULL},
   {"dirlist_cmd",     P_STRING, PI_TEXT,  (void*)&DirBufferCommand,CMT_DIRLIST_CMD,NULL},
@@ -190,12 +201,12 @@ static struct param_ptr params2[] = {
   {"form_color",      P_INT,    PI_SEL_C, (void*)&form_color,CMT_F_COLOR,colorstr},
 #ifdef BG_COLOR
   {"bg_color",        P_INT,    PI_SEL_C, (void*)&bg_color,CMT_BG_COLOR,colorstr},
-#endif
+#endif /* BG_COLOR */
   {"active_style",    P_INT,    PI_ONOFF, (void*)&useActiveColor, CMT_ACTIVE_STYLE, NULL},
   {"active_color",    P_INT,    PI_SEL_C, (void*)&active_color,CMT_C_COLOR,colorstr},
   {NULL, 0, 0, NULL, NULL, NULL},
 };
-#endif
+#endif /* COLOR */
 
 
 static struct param_ptr params3[] = {
@@ -203,7 +214,7 @@ static struct param_ptr params3[] = {
 #ifdef USE_HISTORY
   {"history",         P_INT,    PI_TEXT,  (void*)&URLHistSize,CMT_HISTSIZE,NULL},
   {"save_hist",       P_INT,    PI_ONOFF, (void*)&URLSaveHist,CMT_SAVEHIST,NULL},
-#endif
+#endif /* USE_HISTORY */
   {"confirm_qq",      P_INT,    PI_ONOFF, (void*)&confirm_on_quit, CMT_CONFIRM_QQ,NULL},
   {"show_lnum",       P_INT,    PI_ONOFF, (void*)&showLineNum, CMT_SHOW_NUM,NULL},
   {"ftppasswd",       P_STRING, PI_TEXT,  (void*)&ftppasswd, CMT_FTPPASS, NULL},
@@ -211,13 +222,17 @@ static struct param_ptr params3[] = {
   {"no_referer",      P_INT,    PI_ONOFF, (void*)&NoSendReferer, CMT_NOSENDREFERER,NULL},
   {"accept_language", P_STRING, PI_TEXT,  (void*)&AcceptLang, CMT_ACCEPTLANG, NULL},
   {"wrap_search",     P_INT,    PI_ONOFF, (void*)&WrapDefault, CMT_WRAP, NULL},
+#ifdef VIEW_UNSEENOBJECTS
+  {"view_unseenobject",   P_INT,    PI_ONOFF, (void*)&view_unseenobject, CMT_VIEW_UNSEENOBJECTS, NULL},
+#endif /* VIEW_UNSEENOBJECTS */
 #ifdef USE_COOKIE
   {"use_cookie",      P_INT,    PI_ONOFF, (void*)&use_cookie, CMT_USECOOKIE, NULL},
-#endif
+  {"accept_cookie",   P_INT,    PI_ONOFF, (void*)&accept_cookie, CMT_ACCEPTCOOKIE, NULL},
+#endif /* USE_COOKIE */
 #ifdef MOUSE
   {"use_mouse",       P_INT,    PI_ONOFF, (void*)&use_mouse, CMT_MOUSE, NULL},
   {"reverse_mouse",   P_INT,    PI_ONOFF, (void*)&reverse_mouse, CMT_REVERSE_MOUSE, NULL},
-#endif
+#endif /* MOUSE */
   {NULL, 0, 0, NULL, NULL, NULL},
 };
 
@@ -229,7 +244,7 @@ static struct param_ptr params4[] = {
   {"noproxy_netaddr", P_INT,	PI_ONOFF, (void*)&NOproxy_netaddr, CMT_NOPROXY_NETADDR, NULL },
 #ifdef INET6
   {"dns_order",	      P_INT, PI_SEL_C,  (void*)&DNS_order,CMT_DNS_ORDER, dnsorders},
-#endif
+#endif /* INET6 */
   {NULL, 0, 0, NULL, NULL, NULL},
 };
 
@@ -246,7 +261,9 @@ static struct param_ptr params6[] = {
   {"extbrowser",      P_STRING, PI_TEXT,  (void*)&ExtBrowser,CMT_EXTBRZ},
   {"extbrowser2",     P_STRING, PI_TEXT,  (void*)&ExtBrowser2,CMT_EXTBRZ2},
   {"extbrowser3",     P_STRING, PI_TEXT,  (void*)&ExtBrowser3,CMT_EXTBRZ3},
+#ifndef __EMX__
   {"bgextviewer",     P_INT,    PI_ONOFF, (void*)&BackgroundExtViewer,CMT_BGEXTVIEW},
+#endif /* __EMX__ */
   {NULL,              0,        0, NULL, NULL, NULL},
 };
 
@@ -256,21 +273,21 @@ static struct param_section sections[] =
   {"ɽط", params1},
 #ifdef COLOR
   {"ɽ", params2},
-#endif
+#endif /* COLOR */
   {"¾",params3},
   {"ץ",params4},
   {"ǥ쥯ȥ", params5},
   {"ץ",params6},
-#else
+#else /* LANG != JA */
   {"Display", params1},
 #ifdef COLOR
   {"Color Setting", params2},
-#endif
+#endif /* COLOR */
   {"Other Behavior",params3},
   {"Proxy Setting",params4},
   {"Directory Setting", params5},
   {"External Programs",params6},
-#endif
+#endif /* LANG != JA */
   {NULL,NULL}
 };  
 
@@ -352,6 +369,71 @@ parse_proxy()
   if (NO_proxy)      set_no_proxy(NO_proxy);
 }
 
+#ifdef __EMX__
+static int do_mkdir(const char*dir,long mode){
+  char*r,abs[_MAX_PATH];
+
+  _abspath(abs,rc_dir,_MAX_PATH);	// Translate '\\' to '/'
+  r=strrchr(abs,'/');
+  if(r&&r>abs+2)
+    *r=0;
+  return mkdir(abs,mode);
+}
+#else /* not __EMX__ */
+#define do_mkdir(dir,mode) mkdir(dir,mode)
+#endif /* not __EMX__ */
+
+struct table2*
+loadMimeTypes(char *filename)
+{
+  FILE *f;
+  char *d, *type;
+  int i,n;
+  Str tmp;
+  struct table2 *mtypes;
+  
+  f = fopen(expandName(filename),"r");
+  if (f == NULL) return NULL;
+  n = 0;
+  while (tmp = Strfgets(f),tmp->length > 0) {
+    d = tmp->ptr;
+    if (d[0] != '#') {
+      d = strtok(d, " \t\n\r");
+      d = strtok(NULL, " \t\n\r");
+      for (i=0; d!=NULL; i++)
+        d = strtok(NULL, " \t\n\r");
+      n += i;
+    }
+  }
+  fseek(f,0,0);
+  mtypes = New_N(struct table2,n+1);
+  i = 0;
+  while (tmp = Strfgets(f),tmp->length > 0) {
+    d = tmp->ptr;
+    if (d[0] == '#')
+      continue;
+    type = strtok(d, " \t\n\r");
+    while (1) {
+      d = strtok(NULL, " \t\n\r");
+      if (d == NULL) break;
+      mtypes[i].item1 = Strnew_charp(d)->ptr;
+      mtypes[i].item2 = Strnew_charp(type)->ptr;
+      i++;
+    }
+  }
+  mtypes[i].item1 = NULL;
+  mtypes[i].item2 = NULL;
+  fclose(f);
+  return mtypes;
+}  
+
+void
+initMimeTypes()
+{
+  UserMimeTypes = loadMimeTypes(USER_MIMETYPES);
+  SysMimeTypes = loadMimeTypes(SYS_MIMETYPES);
+}
+
 void
 init_rc(char *config_file)
 {
@@ -365,11 +447,14 @@ init_rc(char *config_file)
 
   if (stat(rc_dir,&st) < 0) {
     if (errno == ENOENT) { /* no directory */
-      if (mkdir(rc_dir,0700) < 0) {
+      if (do_mkdir(rc_dir,0700) < 0) {
 	fprintf(stderr,"Can't create config directory (%s)!",rc_dir);
 	rc_dir = tmpdir;
 	return;
       }
+      else {
+	stat(rc_dir,&st);
+      }
     }
     else {
       fprintf(stderr,"Can't open config directory (%s)!",rc_dir);
@@ -398,14 +483,19 @@ init_rc(char *config_file)
   parse_proxy();
   WrapSearch = WrapDefault;
   initMailcap();
+  initMimeTypes();
 }
 
 
 static char optionpanel_src1[] =
 "<html><head><title>Option Setting Panel</title></head>\
-<body><center><b>Option Setting Panel</b></center><p>\
-<a href=\"file:///$LIB/w3mhelperpanel?mode=panel\">%s</a>\
-<form method=internal action=option>";
+<body><center><b>Option Setting Panel</b></center><p>\n"
+#ifdef __EMX__
+"<a href=\"file:///$LIB/w3mhelperpanel.exe?mode=panel\">%s</a>\n"
+#else /* not __EMX__ */
+"<a href=\"file:///$LIB/w3mhelperpanel?mode=panel\">%s</a>\n"
+#endif /* not __EMX__ */
+"<form method=internal action=option>";
 
 static Str
 to_str(struct param_ptr *p)
@@ -504,7 +594,7 @@ panel_set_option(struct parsed_tagarg* arg)
   parse_proxy();
 #ifdef USE_HISTORY
   resizeHist(URLHist, URLHistSize + 1);
-#endif
+#endif /* USE_HISTORY */
   WrapSearch = WrapDefault;
   backBf();
 }
@@ -523,7 +613,11 @@ rcFile(char *base)
 char *
 libFile(char *base)
 {
+#ifdef __EMX__
+  Str file = Strnew_charp(get_os2_dft("W3M_LIB_DIR", LIB_DIR));
+#else
   Str file = Strnew_charp(LIB_DIR);
+#endif /* __EMX__ */
 
   Strcat_char(file, '/');
   Strcat_charp(file, base);
@@ -533,8 +627,11 @@ libFile(char *base)
 char *
 helpFile(char *base)
 {
+#ifdef __EMX__
+    Str file = Strnew_charp(get_os2_dft("W3M_HELP_DIR",HELP_DIR));
+#else /* not __EMX__ */
     Str file = Strnew_charp(HELP_DIR);
-    
+#endif /* not __EMX__ */
     Strcat_char(file, '/');
     Strcat_charp(file, base);
     return expandName(file->ptr);
