@@ -3,7 +3,7 @@
 
 #ifndef NULL
 #define NULL	0
-#endif
+#endif /* not NULL */
 
 #define	RE_NORMAL	0
 #define RE_MATCHMODE	0x07
