@@ -1,4 +1,5 @@
-/* $Id: table.c,v 1.9 2000/01/21 07:58:41 aito Exp $ */
+/* -*- mode: c; c-basic-offset: 2; tab-width: 8; -*- */
+/* $Id: table.c,v 1.12 2000/04/06 01:25:00 aito Exp $ */
 /*
  * HTML table
  */
@@ -6,6 +7,9 @@
 #include <stdio.h>
 #include <string.h>
 #include <math.h>
+#ifdef __EMX__
+#include <strings.h>
+#endif  /* __EMX__ */
 
 #include "fm.h"
 #include "html.h"
@@ -19,7 +23,7 @@ static char    *ruleB[] = {"00", "", "", "", "", "", "", "07", "", "", "", "0B",
 #define TN_VERTICALBAR ""
 #define HORIZONTALBAR ""
 #define RULE_WIDTH 2
-#else
+#else /* not KANJI_SYMBOLS */
 static char    *rule[] = {
   "<_RULE TYPE=0>+</_RULE>", 
   "<_RULE TYPE=1>+</_RULE>", 
@@ -41,7 +45,7 @@ static char    **ruleB = rule;
 #define TN_VERTICALBAR "<_RULE TYPE=5>|</_RULE>"
 #define HORIZONTALBAR "<_RULE TYPE=10>-</_RULE>"
 #define RULE_WIDTH 1
-#endif
+#endif /* not KANJI_SYMBOLS */
 
 #define RULE(mode) (((mode)==BORDER_THICK)?ruleB:rule)
 #define TK_VERTICALBAR(mode) (RULE(mode)[5])
@@ -60,29 +64,29 @@ static char    **ruleB = rule;
 #define HTT_TRSET  0x40
 #ifdef NOWRAP
 #define HTT_NOWRAP  4
-#endif
+#endif /* NOWRAP */
 #define TAG_IS(s,tag,len) (strncasecmp(s,tag,len)==0&&(s[len] == '>' || IS_SPACE((int)s[len])))
 
 #ifndef max
 #define max(a,b)        ((a) > (b) ? (a) : (b))
-#endif
+#endif /* not max */
 #ifndef min
 #define min(a,b)        ((a) > (b) ? (b) : (a))
-#endif
+#endif /* not min */
 #ifndef abs
 #define abs(a)          ((a) >= 0. ? (a) : -(a))
-#endif
+#endif /* not abs */
 
 #ifdef MATRIX
 #ifndef MESCHACH
 #include "matrix.c"
-#endif
-#endif
+#endif /* not MESCHACH */
+#endif /* MATRIX */
 
 #ifdef MATRIX
 int correct_table_matrix(struct table *, int, int, int, double);
 void set_table_matrix(struct table *, int);
-#endif
+#endif /* MATRIX */
 
 #ifdef MATRIX
 static double 
@@ -105,7 +109,7 @@ weight2(int a)
 #define sigma_td_nw(a)    (32*weight2(a))      /* <td ...> */
 #define sigma_table(a)    (0.25*weight2(a))    /* <table width=...> */
 #define sigma_table_nw(a) (2*weight2(a))       /* <table...> */
-#else
+#else /* not MATRIX */
 static double
 weight3(int x)
 {
@@ -116,7 +120,7 @@ weight3(int x)
   else
     return LOG_MIN*(log((double)x/LOG_MIN) + 1.);
 }
-#endif
+#endif /* not MATRIX */
 
 static int
 dsort_index(short e1, short *ent1, short e2, short *ent2, int base, 
@@ -263,11 +267,19 @@ newTable()
   t->tabdata = New_N(TextList**,MAXROW);
   t->tabattr = New_N(table_attr*,MAXROW);
   t->tabheight = NewAtom_N(short,MAXROW);
+#ifdef ID_EXT
+  t->tabidvalue = New_N(Str*,MAXROW);
+  t->tridvalue = New_N(Str,MAXROW);
+#endif /* ID_EXT */
 
   for (i = 0; i < MAXROW; i++) {
     t->tabdata[i] = NULL;
     t->tabattr[i] = 0;
     t->tabheight[i] = 0;
+#ifdef ID_EXT
+    t->tabidvalue[i] = NULL;
+    t->tridvalue[i] = NULL;
+#endif /* ID_EXT */
   }
   for (j = 0; j < MAXCOL; j++) {
     t->tabwidth[j] = 0;
@@ -284,9 +296,10 @@ newTable()
 #ifdef MATRIX
   t->matrix = NULL;
   t->vector = NULL;
-#endif
-  t->prev_ctype = PC_ASCII;
-  t->prev_spaces = -1;
+#endif /* MATRIX */
+  t->linfo.prev_ctype = PC_ASCII;
+  t->linfo.prev_spaces = -1;
+  t->linfo.prevchar = ' ';
   t->trattr = 0;
 
   t->status = R_ST_NORMAL;
@@ -302,34 +315,60 @@ check_row(struct table *t, int row)
   TextList ***tabdata; 
   table_attr **tabattr;
   short *tabheight;
+#ifdef ID_EXT
+  Str **tabidvalue;
+  Str *tridvalue;
+#endif /* ID_EXT */
 
   if (row >= t->max_rowsize) {
     r = max(t->max_rowsize*2, row + 1);
     tabdata = New_N(TextList**,r);
     tabattr = New_N(table_attr*,r);
     tabheight = New_N(short,r);
+#ifdef ID_EXT
+    tabidvalue = New_N(Str*,r);
+    tridvalue = New_N(Str, r);
+#endif /* ID_EXT */
     for (i = 0; i < t->max_rowsize; i++) {
       tabdata[i] = t->tabdata[i];
       tabattr[i] = t->tabattr[i];
       tabheight[i] = t->tabheight[i];
+#ifdef ID_EXT
+      tabidvalue[i] = t->tabidvalue[i];
+      tridvalue[i] = t->tridvalue[i];
+#endif /* ID_EXT */
     }
     for (; i < r; i++) {
       tabdata[i] = NULL;
       tabattr[i] = NULL;
       tabheight[i] = 0;
+#ifdef ID_EXT
+      tabidvalue[i] = NULL;
+      tridvalue[i] = NULL;
+#endif /* ID_EXT */
     }
     t->tabdata = tabdata;
     t->tabattr = tabattr;
     t->tabheight = tabheight;
+#ifdef ID_EXT
+    t->tabidvalue = tabidvalue;
+    t->tridvalue = tridvalue;
+#endif /* ID_EXT */
     t->max_rowsize = r;
   }
     
   if (t->tabdata[row] == NULL) {
     t->tabdata[row] = New_N(TextList*,MAXCOL);
     t->tabattr[row] = NewAtom_N(table_attr,MAXCOL);
+#ifdef ID_EXT
+    t->tabidvalue[row] = New_N(Str,MAXCOL);
+#endif /* ID_EXT */
     for (i = 0; i < MAXCOL; i++) {
       t->tabdata[row][i] = NULL;
       t->tabattr[row][i] = 0;
+#ifdef ID_EXT
+      t->tabidvalue[row][i] = NULL;
+#endif /* ID_EXT */
     }
   }
 }
@@ -780,11 +819,11 @@ check_maximum_width(struct table *t)
       cell->necell++;
     }
   }
-#else
+#else /* not MATRIX */
   check_cell_width(t->tabwidth, cell->width, cell->col, cell->colspan,
 		   cell->maxcell, cell->index, t->cellspacing, 0);
   check_minimum_width(t, t->tabwidth);
-#endif
+#endif /* not MATRIX */ 
 }
 
 
@@ -896,10 +935,10 @@ check_table_width(struct table *t, short *newwidth, MAT *minv, int itr)
 #ifdef __GNUC__
   short orgwidth[t->maxcol +1];
   short cwidth[cell->maxcell +1], swidth[cell->maxcell +1];
-#else
+#else /* __GNUC__ */
   short orgwidth[MAXCOL];
   short cwidth[MAXCELL], swidth[MAXCELL];
-#endif
+#endif /* __GNUC__ */
   int twidth;
   double sxx, sxy, *Sxx, stotal;
     
@@ -1017,7 +1056,7 @@ check_table_width(struct table *t, short *newwidth, MAT *minv, int itr)
     return corr;
 }
 
-#else
+#else /* not MATRIX */
 void
 set_table_width(struct table *t, short *newwidth, int maxwidth)
 {
@@ -1133,7 +1172,7 @@ set_table_width(struct table *t, short *newwidth, int maxwidth)
     }
   } while (try_again);
 }
-#endif
+#endif /* not MATRIX */
 
 void
 check_table_height(struct table *t)
@@ -1214,9 +1253,9 @@ get_table_width(struct table *t, short *orgwidth, short *cellwidth,
 {
 #ifdef __GNUC__
   short newwidth[t->maxcol +1];
-#else
+#else /* not __GNUC__ */
   short newwidth[MAXCOL];
-#endif
+#endif /* not __GNUC__ */
   int i;
   int swidth;
   struct table_cell *cell = &t->cell;
@@ -1255,9 +1294,9 @@ renderCoTable(struct table * tbl)
   int width, nwidth, indent, maxwidth;
 #ifdef __GNUC__
   short newwidth[tbl->maxcol +1];
-#else
+#else /* not __GNUC__ */
   short newwidth[MAXCOL];
-#endif
+#endif /* not __GNUC__ */
 
   for (i=0; i <= tbl->maxcol; i++)
     newwidth[i] = tbl->tabwidth[i];
@@ -1285,9 +1324,11 @@ renderCoTable(struct table * tbl)
     h_env.limit += b_width;
 #ifndef TABLE_EXPAND
     if (t->total_width == 0)
-	maxwidth = h_env.limit -indent;
+      maxwidth = h_env.limit -indent;
+    else if (t->total_width > 0)
+      maxwidth = t->total_width;
     else
-	maxwidth = t->total_width;
+      maxwidth = t->total_width = - t->total_width * newwidth[col] / 100;
 #else
     maxwidth = h_env.limit -indent;
     if (t->total_width > 0) {
@@ -1304,7 +1345,7 @@ renderCoTable(struct table * tbl)
       }
       t->total_width = maxwidth;
     }
-#endif
+#endif /* TABLE_EXPAND */
     renderTable(t, maxwidth, &h_env);
     width = t->total_width -b_width +indent;
     if (width > nwidth) {
@@ -1372,9 +1413,12 @@ renderTable(struct table * t,
   VEC             *newwidth;
   MAT             *mat, *minv;
   PERM            *pivot;
-#endif
+#endif /* MATRIX */
   int             maxheight = 0;
   Str             vrulea, vruleb, vrulec;
+#ifdef ID_EXT
+  Str		  idtag;
+#endif /* ID_EXT */
 
   if (t->maxcol < 0) {
       make_caption(t, h_env);
@@ -1420,7 +1464,7 @@ renderTable(struct table * t,
           for(i=0;i<=t->maxcol;i++)
             fprintf(stderr,"%2d ",new_tabwidth[i]);
           fprintf(stderr,"\n");
-#endif
+#endif /* TABLE_DEBUG */
           itr++;
 
       } while (check_table_width(t, new_tabwidth, minv, itr));
@@ -1434,12 +1478,12 @@ renderTable(struct table * t,
           t->tabwidth[i] = new_tabwidth[i];
       }
   }
-#else
+#else /* not MATRIX */
   set_table_width(t, new_tabwidth, max_width);
   for (i = 0; i <= t->maxcol; i++) {
       t->tabwidth[i] = new_tabwidth[i];
   }
-#endif
+#endif /* not MATRIX */
   
   renderCoTable(t);
   check_minimum_width(t, t->tabwidth);
@@ -1505,8 +1549,20 @@ renderTable(struct table * t,
       renderbuf = Strnew();
       if (t->border_mode == BORDER_THIN || t->border_mode == BORDER_THICK)
         Strcat(renderbuf,vrulea);
+#ifdef ID_EXT
+	if (t->tridvalue[r] != NULL && h == 0) {
+	  idtag = Sprintf("<a name=\"%s\">", (t->tridvalue[r])->ptr);
+	  Strcat(renderbuf, idtag);
+	}
+#endif /* ID_EXT */
       for (i = 0; i <= t->maxcol; i++) {
         check_row(t,r);
+#ifdef ID_EXT
+	if (t->tabidvalue[r][i] != NULL && h == 0) {
+	  idtag = Sprintf("<a name=\"%s\">", (t->tabidvalue[r][i])->ptr);
+	  Strcat(renderbuf, idtag);
+	}
+#endif /* ID_EXT */
         if (!(t->tabattr[r][i] & HTT_X)) {
           w = t->tabwidth[i];
           for (j = i + 1;
@@ -1637,8 +1693,8 @@ setwidth0(struct table * t, struct table_mode *mode)
   if (t->tabwidth[t->col] < 0)
     return -1;
   check_row(t,t->row);
-  if (t->prev_spaces > 0)
-      width -= t->prev_spaces;
+  if (t->linfo.prev_spaces > 0)
+      width -= t->linfo.prev_spaces;
   w = table_colspan(t, t->row, t->col);
   if (w == 1) {
       if (t->tabwidth[t->col] < width)
@@ -1659,7 +1715,7 @@ setwidth(struct table * t, struct table_mode *mode)
 #ifdef NOWRAP
   if (t->tabattr[t->row][t->col] & HTT_NOWRAP)
       check_minimum0(t, width);
-#endif
+#endif /* NOWRAP */
   if (mode->pre_mode & (TBLM_NOBR|TBLM_PRE|TBLM_PRE_INT) &&
       mode->nobr_offset >= 0)
       check_minimum0(t, width -mode->nobr_offset);
@@ -1680,7 +1736,8 @@ static void
 clearcontentssize(struct table * t, struct table_mode *mode)
 {
   mode->nobr_offset = 0;
-  t->prev_spaces = -1;
+  t->linfo.prev_spaces = -1;
+  t->linfo.prevchar = ' ';
   t->tabcontentssize = 0;
 }
 
@@ -1717,50 +1774,77 @@ check_rowcol(struct table * tbl)
 }
 
 int
-minimum_length(char *line)
+skip_space(struct table *t, char *line, struct table_linfo *linfo,
+	   int checkminimum)
 {
-  Lineprop ctype = PC_ASCII, prev_ctype;
-  char prevchar = '\0';
-  int w;
+  int skip = 0, s = linfo->prev_spaces;
+  Lineprop ctype = linfo->prev_ctype, prev_ctype;
+  char prevchar = linfo->prevchar;
+  int w = (linfo->prev_spaces == -1)? 0 : linfo->length;
   int min = 1;
-
-  if (*line == '<') /* tag */
-    return visible_length(line);
   
-  w = 0;
-  while (*line) {
-      prev_ctype = ctype;
-      ctype = get_ctype(*line, prev_ctype);
-      if (*line == '&') {
-          if (min < w)
-              min = w;
-          w = strlen(getescapecmd(&line));
-      } else {
-          if (ctype==PC_ASCII && *line==' ') {
-              if (min < w)
-                  min = w;
-              w = 0;
-          } else if (prevchar &&
-                     is_boundary(prevchar,prev_ctype, *line,ctype)) {
-              if (min < w)
-                  min = w;
-        w = 1;
-          } else
-              w++;
-    line++;
-  }      
-      prevchar = *(line-1);
+  if (*line=='<' && line[strlen(line)-1]=='>') {
+    if (checkminimum)
+      check_minimum0(t, visible_length(line));
+    return 0;
   }
-  if (min < w)
+  
+  while (*line) {
+    char c = *line, *save = line;
+    int ec = '\0', len = 1;
+    prev_ctype = ctype;
+    ctype = get_ctype(c, prev_ctype);
+    if (min < w)
       min = w;
-  return min;
-}
+    if (ctype==PC_ASCII && IS_SPACE(c)) {
+      w = 0;
+      s++;
+    } else {
+      if (c == '&') {
+	ec = getescapechar(&line);
+	if (ec) {
+	  c = ec;
+	  if (IS_CNTRL(ec))
+	    ctype = PC_CTRL;
+	  else
+	    ctype = PC_ASCII;
+	  len = strlen(conv_latin1(ec));
+	}
+      }
+      if (prevchar && is_boundary(prevchar,prev_ctype, c,ctype)) {
+	w = len;
+      } else {
+	w += len;
+      }
+      if (s > 0) {
+#ifdef JP_CHARSET
+	if (ctype==PC_KANJI1 && prev_ctype==PC_KANJI2)
+	  skip += s;
+	else
+#endif /* JP_CHARSET */
+	  skip += s -1;
+      }
+      s = 0;
+    }      
+    prevchar = c;
+    line = save +len;
+  }
+  if (s > 1) {
+    skip += s -1;
+    linfo->prev_spaces = 1;
+  } else {
+    linfo->prev_spaces = s;
+  }
+  linfo->prev_ctype = ctype;
+  linfo->prevchar = prevchar;
 
-
-void
-check_minimum(struct table *t, char *line)
-{
-  check_minimum0(t,minimum_length(line));
+  if (checkminimum) {
+    if (min < w)
+      min = w;
+    linfo->length = w;
+    check_minimum0(t, min);
+  }
+  return skip;
 }
 
 #define TAG_ACTION_NONE 0
@@ -1806,6 +1890,11 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
         else if (!strcasecmp(t->value, "center"))
           align = (HTT_CENTER | HTT_TRSET);
       }
+#ifdef ID_EXT
+      if (!strcasecmp(t->arg, "id") && t->value) {
+	tbl->tridvalue[tbl->row] = Strnew_charp(t->value);
+      }
+#endif /* ID_EXT */
     }
     tbl->trattr = align;
     break;
@@ -1821,8 +1910,9 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
     mode->mode_level = 0;
     mode->pre_mode = 0;
     tbl->flag |= TBL_IN_COL;
-    tbl->prev_spaces = -1;
-    tbl->prev_ctype = PC_ASCII;
+    tbl->linfo.prev_spaces = -1;
+    tbl->linfo.prevchar = ' ';
+    tbl->linfo.prev_ctype = PC_ASCII;
     tbl->indent = 0;
     if (tbl->row == -1) {
       /* for broken HTML... */
@@ -1880,7 +1970,7 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
 #ifdef NOWRAP
       else if (!strcasecmp(t->arg, "nowrap"))
         tbl->tabattr[tbl->row][tbl->col] |= HTT_NOWRAP;
-#endif
+#endif /* NOWRAP */
       else if (!strcasecmp(t->arg,"width") && t->value) {
 	if (IS_DIGIT(*t->value)) {
 	  v = atoi(t->value);
@@ -1893,12 +1983,16 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
 	  } else {
 #ifdef TABLE_EXPAND
 	    v = max(v/tbl->ppc,1);
-#else
+#else /* not TABLE_EXPAND */
 	    v = v/PIXEL_PER_CHAR;
-#endif
+#endif /* not TABLE_EXPAND */
 	  }
 	} else 
           continue;
+#ifdef ID_EXT
+      } else if (!strcasecmp(t->arg, "id") && t->value) {
+	tbl->tabidvalue[tbl->row][tbl->col] = Strnew_charp(t->value);
+#endif /* ID_EXT */
       }
     }
     tbl->tabattr[tbl->row][tbl->col] &= ~HTT_ALIGN;
@@ -1935,7 +2029,7 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
         if (v0 == 0 || (v0 > 0 && v > v0) || (v0 < 0 && v < v0)) {
 #ifdef TABLE_DEBUG
           fprintf(stderr,"width(%d) = %d\n",tbl->col,v);
-#endif
+#endif /* TABLE_DEBUG */
           tbl->fixed_width[tbl->col] = v;
         }
       } else if (cell->icell >= 0) {
@@ -1971,6 +2065,7 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
     return TAG_ACTION_NONE;
   case HTML_P: case HTML_BR: case HTML_DT: case HTML_DD:
   case HTML_CENTER: case HTML_N_CENTER: case HTML_DIV: case HTML_N_DIV:
+  case HTML_H: case HTML_N_H:
     check_rowcol(tbl);
     pushdata(tbl, tbl->row, tbl->col, line);
     setwidth(tbl, mode);
@@ -1984,16 +2079,18 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
     setwidth(tbl, mode);
     clearcontentssize(tbl, mode);
     mode->indent_level++;
-    tbl->indent += INDENT_INCR;
+    if (mode->indent_level <= MAX_INDENT_LEVEL)
+      tbl->indent += INDENT_INCR;
     break;
   case HTML_N_DL: case HTML_N_BLQ: case HTML_N_OL: case HTML_N_UL:
+    if (mode->indent_level > 0) {
     check_rowcol(tbl);
     pushdata(tbl, tbl->row, tbl->col, line);
     setwidth(tbl, mode);
     clearcontentssize(tbl, mode);
-    if (mode->indent_level > 0) {
       mode->indent_level--;
-      tbl->indent -= INDENT_INCR;
+      if (mode->indent_level < MAX_INDENT_LEVEL)
+        tbl->indent -= INDENT_INCR;
     }
     break;
   case HTML_LI:
@@ -2021,8 +2118,9 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
       mode->mode_level--;
     if (mode->mode_level == 0)
       mode->pre_mode &= ~TBLM_PRE;
-    tbl->prev_spaces = 0;
-    tbl->prev_ctype = PC_ASCII;
+    tbl->linfo.prev_spaces = 0;
+    tbl->linfo.prevchar = '\0';
+    tbl->linfo.prev_ctype = PC_ASCII;
     break;
   case HTML_NOBR:
     check_rowcol(tbl);
@@ -2053,7 +2151,7 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
       mode->nobr_offset = -1;
       mode->pre_mode |= TBLM_PRE_INT;
     }
-    tbl->prev_spaces = 0;
+    tbl->linfo.prev_spaces = 0;
     break;
   case HTML_N_PRE_INT:
     check_rowcol(tbl);
@@ -2071,7 +2169,7 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
   case HTML_N_FORM:
     process_n_form();
     break;
-#else
+#else /* not NEW_FORM */
   case HTML_FORM:
   case HTML_N_FORM:
     check_rowcol(tbl);
@@ -2088,7 +2186,7 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
     }
     pushdata(tbl, tbl->row, tbl->col, tok->ptr);
     break;
-#endif
+#endif /* not NEW_FORM */
   case HTML_INPUT:
     tmp = process_input(parse_tag(line + 6));
     if (tmp)
@@ -2102,12 +2200,12 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
 #ifdef MENU_SELECT
     if (!tag_exists(t_arg,"multiple")) /* non-multiple select */
       mode->pre_mode |= TBLM_INSELECT;
-#endif
+#endif /* MENU_SELECT */
     break;
   case HTML_N_SELECT:
 #ifdef MENU_SELECT
     mode->pre_mode &= ~TBLM_INSELECT;
-#endif
+#endif /* MENU_SELECT */
     tmp = process_n_select();
     if (tmp)
       feed_table(tbl,tmp->ptr,mode,width);
@@ -2119,7 +2217,7 @@ feed_table_tag(struct table * tbl, char *line, struct table_mode *mode, int widt
 #ifdef MENU_SELECT
     else
       feed_select(line);
-#endif
+#endif /* MENU_SELECT */
     break;
   case HTML_TEXTAREA:
     w = 0;
@@ -2228,6 +2326,7 @@ feed_table(struct table * tbl, char *tline, struct table_mode *mode, int width)
   char           *line;
   char           *p;
   Str             tok, tmp;
+  struct table_linfo *linfo = &tbl->linfo;
 
   if (tline == NULL || *tline == '\0')
     return;
@@ -2268,82 +2367,56 @@ feed_table(struct table * tbl, char *tline, struct table_mode *mode, int width)
       feed_select(line);
       continue;
     }
-#endif
-    if (!(mode->pre_mode & (TBLM_PRE|TBLM_PRE_INT))) {
-      Lineprop ctype;
-      if (!(tbl->flag & TBL_IN_COL)) {
-        tbl->prev_spaces = -1;
-        tbl->prev_ctype = PC_ASCII;
+#endif /* MENU_SELECT */
+    /* convert &...; if it is latin-1 character */
+    if (!(*line=='<' && line[strlen(line)-1]=='>') &&
+	strchr(line,'&') != NULL) {
+      tmp = Strnew();
+      for (p = line; *p;) {
+	char *q,*r;
+	if (*p == '&') {
+	  if (!strncasecmp(p,"&amp;",5) ||
+	      !strncasecmp(p,"&gt;",4) ||
+	      !strncasecmp(p,"&lt;",4)) {
+	    /* do not convert */
+	    Strcat_char(tmp,*p);
+	    p++;
+	  }
+	  else {
+	    q = p;
+	    r = getescapecmd(&p);
+	    if (r != NULL && ((*r & 0x80) || IS_CNTRL(*r))) {
+	      /* latin-1 character */
+	      Strcat_charp(tmp,r);
+	    }
+	    else {
+	      Strcat_char(tmp,*q);
+	      p = q+1;
+	    }
+	  }
+	}
+	else {
+	  Strcat_char(tmp,*p);
+	  p++;
+	}
       }
-      if (tbl->prev_spaces != 0)
+      line = tmp->ptr;
+    }
+    if (!(mode->pre_mode & (TBLM_PRE|TBLM_PRE_INT))) {
+      if (!(tbl->flag & TBL_IN_COL)) {
+        linfo->prev_spaces = -1;
+        linfo->prev_ctype = PC_ASCII;
+      }
+      if (linfo->prev_spaces != 0)
         while (IS_SPACE(*line)) line++;
       if (*line == '\0')
         continue;
       check_rowcol(tbl);
       if (mode->pre_mode & TBLM_NOBR && mode->nobr_offset < 0)
         mode->nobr_offset = tbl->tabcontentssize;
-      i = 0;
-      if (*line=='<' && line[strlen(line)-1]=='>')
-        goto _last;
-      /* convert &...; if it is latin-1 character */
-      if (strchr(line,'&') != NULL) {
-        tmp = Strnew();
-        for (p = line; *p;) {
-          char *q,*r;
-          if (*p == '&') {
-            if (!strncasecmp(p,"&amp;",5) ||
-                !strncasecmp(p,"&gt;",4) ||
-                !strncasecmp(p,"&lt;",4)) {
-              /* do not convert */
-              Strcat_char(tmp,*p);
-              p++;
-            }
-            else {
-              q = p;
-              r = getescapecmd(&p);
-              if (r != NULL && ((*r & 0x80) || IS_CNTRL(*r))) {
-                /* latin-1 character */
-                Strcat_charp(tmp,r);
-              }
-              else {
-                Strcat_char(tmp,*q);
-                p = q+1;
-              }
-            }
-          }
-          else {
-            Strcat_char(tmp,*p);
-            p++;
-          }
-        }
-        line = tmp->ptr;
-      }
+
       /* count of number of spaces skipped in normal mode */
-      for (p=line; *p; p++) {
-        char c = *p;
-        if (!IS_SPACE(c)) {
-          ctype = get_ctype(c, tbl->prev_ctype);
-          if (tbl->prev_spaces > 0) {
-#ifdef JP_CHARSET
-            if (ctype==PC_KANJI1 && tbl->prev_ctype==PC_KANJI2)
-              i += tbl->prev_spaces;
-            else
-#endif
-              i += tbl->prev_spaces -1;
-          }
-          tbl->prev_spaces = 0;
-          tbl->prev_ctype = ctype;
-        } else if (tbl->prev_spaces >= 0) {
-          tbl->prev_spaces++;
-        }
-      } /* end for */
-      if (tbl->prev_spaces > 1) {
-        i += tbl->prev_spaces -1;
-        tbl->prev_spaces = 1;
-      }
-    _last:
-      if (!(mode->pre_mode & TBLM_NOBR))
-        check_minimum(tbl, line);
+      i = skip_space(tbl, line, linfo, !(mode->pre_mode & TBLM_NOBR));
       addcontentssize(tbl, visible_length(line) -i);
       setwidth(tbl, mode);
     }
@@ -2382,9 +2455,9 @@ pushTable(struct table * tbl, struct table * tbl1)
         if (tbl->tables)
 #ifdef __CYGWIN__
             bcopy((const char *)tbl->tables, (char *)tmp, (size_t)tbl->ntable*sizeof(struct table_in));
-#else
+#else /* not __CYGWIN__ */
             bcopy(tbl->tables, tmp, tbl->ntable*sizeof(struct table_in));
-#endif /* __CYGWIN__ */
+#endif /* not __CYGWIN__ */
         tbl->tables = tmp;
     }
         
@@ -2478,10 +2551,10 @@ set_table_matrix0(struct table *t, int maxwidth)
 #ifdef __GNUC__
     double we[size];
     char expand[size];
-#else
+#else /* not __GNUC__ */
     double we[MAXCOL];
     char expand[MAXCOL];
-#endif
+#endif /* not __GNUC__ */
     struct table_cell *cell = &t->cell;
 
     w0 = 0.;
@@ -2592,4 +2665,4 @@ set_table_matrix(struct table *t, int width)
     }
     correct_table_matrix(t, 0, size, width, b);
 }
-#endif
+#endif /* MATRIX */
