@@ -1,4 +1,4 @@
-/* $Id: terms.c,v 1.6 2000/01/21 08:32:03 aito Exp $ */
+/* $Id: terms.c,v 1.10 2000/04/06 01:25:00 aito Exp $ */
 /*
  * An original curses library for EUC-kanji by Akinori ITO,     December 1989
  * revised by Akinori ITO, January 1995
@@ -14,14 +14,28 @@
 #include <string.h>
 #ifdef USE_GPM
 #include <gpm.h>
-#endif
+#endif /* USE_GPM */
+#ifdef USE_SYSMOUSE
+#include <machine/console.h>
+#ifndef FBIO_GETMODE    /* FreeBSD 3.x */
+#define FBIO_GETMODE    CONS_GET
+#define FBIO_MODEINFO   CONS_MODEINFO
+#endif /* not FBIO_GETMODE */
+int (*sysm_handler)(int x, int y, int nbs, int obs);
+static int cwidth = 8, cheight = 16;
+static int xpix, ypix, nbs, obs = 0;
+#endif /* USE_SYSMOUSE */
 #ifdef MOUSE
 static int is_xterm = 0;
-#endif
+#endif /* MOUSE */
 
 #ifdef AIX
 #include <sys/select.h>
-#endif
+#endif /* AIX */
+
+#ifdef __EMX__
+#include <sys/select.h>
+#endif /* __EMX__ */
 
 #include "terms.h"
 #include "fm.h"
@@ -34,11 +48,11 @@ void flush_tty();
 #ifdef MOUSE
 void mouse_init(),mouse_end();
 int mouseActive = 0;
-#endif
+#endif /* MOUSE */
 
 #ifndef SIGIOT
 #define SIGIOT SIGABRT
-#endif
+#endif /* not SIGIOT */
 
 #ifdef TERMIO
 #include        <sys/ioctl.h>
@@ -48,7 +62,7 @@ typedef struct termio TerminalMode;
 #define TerminalGet(fd,x)       ioctl(fd,TCGETA,x)
 #define MODEFLAG(d)     ((d).c_lflag)
 #define IMODEFLAG(d)    ((d).c_iflag)
-#endif
+#endif /* TERMIO */
 
 #ifdef TERMIOS
 #include <termios.h>
@@ -58,7 +72,7 @@ typedef struct termios TerminalMode;
 #define TerminalGet(fd,x)       tcgetattr(fd,x)
 #define MODEFLAG(d)     ((d).c_lflag)
 #define IMODEFLAG(d)    ((d).c_iflag)
-#endif
+#endif /* TERMIOS */
 
 #ifdef SGTTY
 #include        <sys/ioctl.h>
@@ -67,7 +81,7 @@ typedef struct sgttyb TerminalMode;
 #define TerminalSet(fd,x)       ioctl(fd,TIOCSETP,x)
 #define TerminalGet(fd,x)       ioctl(fd,TIOCGETP,x)
 #define MODEFLAG(d)     ((d).sg_flags)
-#endif
+#endif /* SGTTY */
 
 #define MAX_LINE        200
 #define MAX_COLUMN      300
@@ -86,7 +100,7 @@ typedef struct sgttyb TerminalMode;
 #ifdef JP_CHARSET
 #define C_WCHAR1        0x40
 #define C_WCHAR2        0x80
-#endif
+#endif /* JP_CHARSET */
 #define C_CTRL          0xc0
 
 #define CHMODE(c)       ((c)&C_WHICHCHAR)
@@ -120,7 +134,7 @@ typedef struct sgttyb TerminalMode;
 #define COL_BTERM       0x0000
 
 #define S_BCOLORED      0xf000
-#endif
+#endif /* BG_COLOR */
 
 
 #define S_GRAPHICS      0x10
@@ -185,7 +199,7 @@ void            clear(),wrap(),touch_line(), touch_column(int);
 #ifdef JP_CHARSET
 void            switch_wchar(FILE*);
 void            switch_ascii(FILE*);
-#endif
+#endif /* JP_CHARSET */
 void            need_clrtoeol(void), clrtoeol(void);
 
 int write1(char);
@@ -208,9 +222,9 @@ set_tty(void)
   else
 #ifndef __EMX__
     ttyn = "/dev/tty";
-#else
+#else /* __EMX__ */
     ttyn = "con";
-#endif
+#endif /* __EMX__ */
   tty = open(ttyn, O_RDWR);
   if (tty < 0) {
     /* use stderr instead of stdin... is it OK???? */
@@ -231,7 +245,7 @@ ttymode_set(int mode,int imode)
   MODEFLAG(ioval) |= mode;
 #ifndef SGTTY
   IMODEFLAG(ioval) |= imode;
-#endif
+#endif /* not SGTTY */
 
   er = TerminalSet(tty, &ioval);
 
@@ -251,7 +265,7 @@ ttymode_reset(int mode,int imode)
   MODEFLAG(ioval) &= ~mode;
 #ifndef SGTTY
   IMODEFLAG(ioval) &= ~imode;
-#endif
+#endif /* not SGTTY */
 
   er = TerminalSet(tty, &ioval);
 
@@ -276,7 +290,7 @@ set_cc(int spec, int val)
     reset_exit(SIGNAL_ARGLIST);
   }
 }
-#endif
+#endif /* not SGTTY */
 
 void
 reset_tty(void)
@@ -304,7 +318,7 @@ reset_exit(SIGNAL_ARG)
 #ifdef MOUSE
   if (mouseActive)
     mouse_end();
-#endif
+#endif /* MOUSE */
   exit(0);
   SIGNAL_RETURN;
 }
@@ -328,7 +342,9 @@ set_int(void)
   signal(SIGILL, error_dump);
   signal(SIGIOT, error_dump);
   signal(SIGFPE, error_dump);
+#ifdef	SIGBUS
   signal(SIGBUS, error_dump);
+#endif /* SIGBUS */
 /*  signal(SIGSEGV, error_dump); */
 }
 
@@ -360,10 +376,22 @@ extern char *allocStr(char*,int);
 void
 getTCstr(void)
 {
+  char           *ent;
   char           *suc;
   char           *pt = funcstr;
+  int            r;
 
-  tgetent(bp, getenv("TERM"));
+#ifdef	__DJGPP__
+  ent = getenv("TERM") ? getenv("TERM") : "dosansi";
+#else
+  ent = getenv("TERM");
+#endif /* __DJGPP__ */
+  r = tgetent(bp, ent);
+  if (r != 1) {
+    /* Can't find termcap entry */
+    fprintf(stderr,"Can't find termcap entry %s\n",ent);
+    reset_exit(SIGNAL_ARGLIST);
+  }
 
   GETSTR(T_ce, "ce");           /* clear to the end of line */
   GETSTR(T_cd, "cd");           /* clear to the end of display */
@@ -371,7 +399,7 @@ getTCstr(void)
   if (suc == NULL)
     GETSTR(T_kr, "kr");
   if (tgetflag("bs"))
-    T_kl = "";                        /* cursor left */
+    T_kl = "\b";		/* cursor left */
   else {
     GETSTR(T_kl, "le");
     if (suc == NULL)
@@ -410,7 +438,7 @@ getTCstr(void)
   T_as[0] = '\0';
   T_ae[0] = '\0';
   T_ac[0] = '\0';
-#endif
+#endif /* CYGWIN */
  
   LINES = COLS = 0;
   setlinescols();
@@ -419,14 +447,29 @@ getTCstr(void)
 
 #ifndef TIOCGWINSZ
 #include <sys/ioctl.h>
-#endif
+#endif /* not TIOCGWINSZ */
 
 void
 setlinescols(void)
 {
   char           *p;
   int             i;
-#if defined(TERMIOS) && defined(TIOCGWINSZ)
+#ifdef __EMX__
+  {
+    int s[2];
+    _scrsize(s);
+    COLS = s[0];
+    LINES = s[1];
+
+    if(getenv("WINDOWID")){
+      FILE*fd=popen("scrsize","rt");
+      if(fd){
+	fscanf(fd,"%i %i",&COLS,&LINES);
+	pclose(fd);
+      }
+    }
+  }
+#elif defined(TERMIOS) && defined(TIOCGWINSZ)
   struct winsize  wins;
 
   i = ioctl(tty, TIOCGWINSZ, &wins);
@@ -434,7 +477,7 @@ setlinescols(void)
     LINES = wins.ws_row;
     COLS = wins.ws_col;
   }
-#endif
+#endif / * defined(TERMIOS) && defined(TIOCGWINSZ) */
   if (LINES <= 0 &&
       (p = getenv("LINES")) != NULL &&
       (i = atoi(p)) >= 0)
@@ -494,7 +537,7 @@ write1(char c)
 {
 #ifdef SCREEN_DEBUG
   usleep(50);
-#endif
+#endif /* SCREEN_DEBUG */
 #ifdef JP_CHARSET
   if (c & 0x80) {
     switch (wmode) {
@@ -525,12 +568,12 @@ write1(char c)
     }
     putc(c, ttyf);
   }
-#else
+#else /* not JP_CHARSET */
   putc(c, ttyf);
-#endif
+#endif /* not JP_CHARSET */
 #ifdef SCREEN_DEBUG
   fflush(ttyf);
-#endif
+#endif /* SCREEN_DEBUG */
   return 0;
 }
 
@@ -595,7 +638,7 @@ putchars(unsigned char c1, unsigned char c2, FILE * f)
     break;
   }
 }
-#endif
+#endif /* JP_CHARSET */
 
 void
 move(int line, int column)
@@ -608,9 +651,9 @@ move(int line, int column)
 
 #ifdef BG_COLOR
 #define M_SPACE (S_SCREENPROP|S_COLORED|S_BCOLORED|S_GRAPHICS)
-#else
+#else /* not BG_COLOR */
 #define M_SPACE (S_SCREENPROP|S_COLORED|S_GRAPHICS)
-#endif
+#endif /* not BG_COLOR */
 
 static int 
 need_redraw(char c1, l_prop pr1, char c2, l_prop pr2)
@@ -648,7 +691,7 @@ addch(char c)
     if (c == ' ' &&
 #ifdef JP_CHARSET
         CHMODE(CurrentMode) != C_WCHAR1 &&
-#endif
+#endif /* JP_CHARSET */
         !(CurrentMode & M_SPACE)) {
       CurColumn++;
       return;
@@ -666,7 +709,7 @@ addch(char c)
     SETCHMODE(CurrentMode, C_WCHAR2);
   else if (c & 0x80)
     SETCHMODE(CurrentMode, C_WCHAR1);
-#endif
+#endif /* JP_CHARSET */
   else if (c & ~0x1f)
     SETCHMODE(CurrentMode, C_ASCII);
   else
@@ -687,7 +730,7 @@ addch(char c)
 #ifdef JP_CHARSET
       else if (CHMODE(pr[CurColumn+1]) == C_WCHAR1 && CurColumn < COLS-2)
         touch_column(CurColumn+2);
-#endif
+#endif /* JP_CHARSET */
     }
   }
 
@@ -713,7 +756,7 @@ addch(char c)
     p = ScreenImage[CurLine]->lineimage;
     pr = ScreenImage[CurLine]->lineprop;
   }
-#endif
+#endif /* JP_CHARSET */
   if (CHMODE(CurrentMode) != C_CTRL) {
     if (need_redraw(p[CurColumn], pr[CurColumn], c, CurrentMode)) {
       p[CurColumn] = c;
@@ -725,7 +768,7 @@ addch(char c)
         touch_column(CurColumn+1);
       else if (CHMODE(CurrentMode) == C_WCHAR2)
         touch_column(CurColumn-1);
-#endif
+#endif /* JP_CHARSET */
     }
     CurColumn++;
   } else if (c == '\t') {
@@ -756,7 +799,7 @@ addch(char c)
     if (CurColumn > 0 &&
         CHMODE(pr[CurColumn]) == C_WCHAR2)
       CurColumn--;
-#endif
+#endif /* JP_CHARSET */
   }
 }
 
@@ -808,7 +851,7 @@ toggle_stand(void)
 #ifdef JP_CHARSET
   if (CHMODE(pr[CurColumn]) == C_WCHAR1)
     pr[CurColumn+1] ^= S_STANDOUT;
-#endif
+#endif /* JP_CHARSET */
 }
 
 void
@@ -853,7 +896,7 @@ graph_ok(void)
 #ifndef KANJI_SYMBOLS
   if (no_graphic_char)
     return 0;
-#endif
+#endif /* not KANJI_SYMBOLS */
   return T_as[0] != 0 && T_ae[0] != 0 && T_ac[0] != 0;
 }
 
@@ -889,16 +932,16 @@ bcolor_seq(int colmode)
   sprintf(seqbuf, "\033[%dm", ((colmode >> 12) & 7) + 40);
   return seqbuf;
 }
-#endif
+#endif /* BG_COLOR */
 
 #define RF_NEED_TO_MOVE    0
 #define RF_CR_OK           1
 #define RF_NONEED_TO_MOVE  2
 #ifdef BG_COLOR
 #define M_MEND (S_STANDOUT|S_UNDERLINE|S_BOLD|S_COLORED|S_BCOLORED|S_GRAPHICS)
-#else
+#else /* not BG_COLOR */
 #define M_MEND (S_STANDOUT|S_UNDERLINE|S_BOLD|S_COLORED|S_GRAPHICS)
-#endif
+#endif /* not BG_COLOR */
 void
 refresh(void)
 {
@@ -910,7 +953,7 @@ refresh(void)
   l_prop          color = COL_FTERM;
 #ifdef BG_COLOR
   l_prop          bcolor = COL_BTERM;
-#endif
+#endif /* BG_COLOR */
   short          *dirty;
 
   for (line = 0; line < LINES; line++) {
@@ -978,19 +1021,19 @@ refresh(void)
 #ifndef BG_COLOR
         if (line == LINES-1 && col == COLS-1)
           break;
-#endif
+#endif /* not BG_COLOR */
         if ((!(pr[col] & S_STANDOUT) && (mode & S_STANDOUT)) ||
             (!(pr[col] & S_UNDERLINE) && (mode & S_UNDERLINE)) ||
             (!(pr[col] & S_BOLD) && (mode & S_BOLD)) ||
             (!(pr[col] & S_COLORED) && (mode & S_COLORED))
 #ifdef BG_COLOR
             || (!(pr[col] & S_BCOLORED) && (mode & S_BCOLORED))
-#endif
+#endif /* BG_COLOR */
             || (!(pr[col] & S_GRAPHICS) && (mode & S_GRAPHICS))) {
           if ((!(pr[col] & S_COLORED) && (mode & S_COLORED))
 #ifdef BG_COLOR
               || (!(pr[col] & S_BCOLORED) && (mode & S_BCOLORED))
-#endif
+#endif /* BG_COLOR */
             )
             writestr(T_op);
           if (!(pr[col] & S_GRAPHICS) && (mode & S_GRAPHICS))
@@ -1016,7 +1059,7 @@ refresh(void)
             mode = (mode & ~COL_BCOLOR | bcolor);
             writestr(bcolor_seq(bcolor));
           }
-#endif
+#endif /* BG_COLOR */
           if ((pr[col] & S_STANDOUT) && !(mode & S_STANDOUT)) {
             writestr(T_so);
             mode |= S_STANDOUT;
@@ -1051,7 +1094,7 @@ refresh(void)
       if (mode & (S_COLORED
 #ifdef BG_COLOR
                   |S_BCOLORED
-#endif
+#endif /* BG_COLOR */
         ))
         writestr(T_op);
       if (mode & S_GRAPHICS)
@@ -1061,7 +1104,7 @@ refresh(void)
     }
 #ifdef JP_CHARSET
     endline();
-#endif
+#endif /* JP_CHARSET */
   }
   MOVE(CurLine, CurColumn);
   fflush(ttyf);
@@ -1152,7 +1195,7 @@ need_clrtoeol(void)
 
   ScreenImage[CurLine]->isdirty |= L_NEED_CE;
 }
-#endif
+#endif /* 0 */
 
 void
 clrtoeol(void)
@@ -1200,14 +1243,14 @@ clrtoeolx(void)
 {
   clrtoeol_with_bcolor();
 }
-#else
+#else /* not BG_COLOR */
 
 void
 clrtoeolx(void)
 {
   clrtoeol();
 }
-#endif
+#endif /* not BG_COLOR */
 
 void
 clrtobot_eol(void (* clrtoeol)())
@@ -1246,7 +1289,7 @@ no_clrtoeol(void)
   
   ScreenImage[CurLine]->isdirty &= ~L_CLRTOEOL;
 }
-#endif
+#endif /* 0 */
 
 void
 addstr(char *s)
@@ -1270,10 +1313,10 @@ addnstr(char *s, int n)
     } else                              /* Ascii or WCHAR2 */
       addch(*s);
   }
-#else
+#else /* not JP_CHARSET */
   for(i = 0; i < n && *s != '\0'; i++)
     addch(*(s++));
-#endif
+#endif /* not JP_CHARSET */
 }
 
 void
@@ -1295,10 +1338,10 @@ addnstr_sup(char *s, int n)
       i++;
     }
   }
-#else
+#else /* not JP_CHARSET */
   for(i = 0; i < n && *s != '\0'; i++)
     addch(*(s++));
-#endif
+#endif /* not JP_CHARSET */ 
   for(; i< n; i++)
     addch(' ');
 }
@@ -1311,15 +1354,15 @@ crmode(void)
   ttymode_set(ISIG,0);
 #ifdef TERMIOS
   set_cc(VMIN, 1);
-#else
+#else /* not TERMIOS */
   set_cc(VEOF, 1);
-#endif
+#endif /* not TERMIOS */
 }
-#else
+#else /* SGTTY */
 {
   ttymode_set(CBREAK,0);
 }
-#endif
+#endif /* SGTTY */
 
 void
 nocrmode(void)
@@ -1328,15 +1371,15 @@ nocrmode(void)
   ttymode_set(ICANON,0);
 #ifdef TERMIOS
   set_cc(VMIN, 4);
-#else
+#else /* not TERMIOS */
   set_cc(VEOF, 4);
-#endif
+#endif /* not TERMIOS */
 }
-#else
+#else /* SGTTY */
 {
   ttymode_reset(CBREAK,0);
 }
-#endif
+#endif /* SGTTY */
 
 void
 term_echo(void) {
@@ -1353,22 +1396,22 @@ term_raw(void)
 #ifndef SGTTY
 #ifdef IEXTEN
 #define TTY_MODE ISIG|ICANON|ECHO|IEXTEN
-#else
+#else /* not IEXTEN */
 #define TTY_MODE ISIG|ICANON|ECHO
-#endif
+#endif /* not IEXTEN */
 {
   ttymode_reset(TTY_MODE,IXON|IXOFF);
 #ifdef TERMIOS
   set_cc(VMIN, 1);
-#else
+#else /* not TERMIOS */
   set_cc(VEOF, 1);
-#endif
+#endif /* not TERMIOS */
 }
-#else
+#else /* SGTTY */
 {
   ttymode_set(RAW,0);
 }
-#endif
+#endif /* SGTTY */
 
 void
 term_cooked(void)
@@ -1377,15 +1420,15 @@ term_cooked(void)
   ttymode_set(TTY_MODE,0);
 #ifdef TERMIOS
   set_cc(VMIN, 4);
-#else
+#else /* not TERMIOS */
   set_cc(VEOF, 4);
-#endif
+#endif /* not TERMIOS */
 }
-#else
+#else /* SGTTY */
 {
   ttymode_reset(RAW,0);
 }
-#endif
+#endif /* SGTTY */
 
 void
 term_cbreak(void)
@@ -1421,7 +1464,57 @@ do_getch()
   else
     return Gpm_Getch();
 }
-#endif
+#endif /* USE_GPM */
+
+#ifdef USE_SYSMOUSE
+int
+sysm_getch()
+{
+  fd_set rfd;
+  int key, x, y;
+
+  FD_ZERO(&rfd);
+  FD_SET(tty, &rfd);
+  while (select(tty+1, &rfd, NULL, NULL, NULL) <= 0) {
+    if (errno == EINTR) {
+      x = xpix / cwidth;
+      y = ypix / cheight;
+      key = (*sysm_handler)(x, y, nbs, obs);
+      if (key != 0)
+	return key;
+    }
+  }
+  return getch();
+}
+
+int
+do_getch()
+{
+  if (is_xterm)
+    return getch();
+  else
+    return sysm_getch();
+}
+
+MySignalHandler
+sysmouse(SIGNAL_ARG)
+{
+  struct mouse_info mi;
+
+  mi.operation = MOUSE_GETINFO;
+  if (ioctl(tty, CONS_MOUSECTL, &mi) == -1)
+    return;
+  xpix = mi.u.data.x;
+  ypix = mi.u.data.y;
+  obs = nbs;
+  nbs = mi.u.data.buttons & 0x7;
+  /* for cosmetic bug in syscons.c on FreeBSD 3.[34] */
+  mi.operation = MOUSE_HIDE;
+  ioctl(tty, CONS_MOUSECTL, &mi);
+  mi.operation = MOUSE_SHOW;
+  ioctl(tty, CONS_MOUSECTL, &mi);
+}
+#endif /* USE_SYSMOUSE */
 
 void
 bell(void)
@@ -1465,26 +1558,50 @@ void mouse_init()
 #ifdef USE_GPM
   Gpm_Connect conn;
   extern int gpm_process_mouse(Gpm_Event*,void*);
+#endif /* USE_GPM */
+#ifdef USE_SYSMOUSE
+  mouse_info_t mi;
+  video_info_t vi;
+  extern int sysm_process_mouse();
+#endif /* USE_SYSMOUSE */
 
-  if (mouseActive) return;
+#if defined(USE_GPM) || defined(USE_SYSMOUSE)
+  if (mouseActive)
+    return;
+#endif /* defined(USE_GPM) || defined(USE_SYSMOUSE) */
   term = getenv("TERM");
   if (!strncmp(term,"kterm",5) || !strncmp(term,"xterm",5)) {
     is_xterm = 1;
   }
+#ifdef USE_GPM
   else {
     conn.eventMask = ~0;
-    conn.defaultMask = GPM_MOVE|GPM_HARD;
+    conn.defaultMask = 0;
     conn.maxMod = 0;
     conn.minMod = 0;
     Gpm_Open(&conn,0); /* don't care even if it fails */
     gpm_handler = gpm_process_mouse;
   }
-#else
-  term = getenv("TERM"); 
-  if (!strncmp(term,"kterm",5) || !strncmp(term,"xterm",5)) {
-    is_xterm = 1;
+#endif /* USE_GPM */
+#ifdef USE_SYSMOUSE
+  else {
+    if (ioctl(tty, FBIO_GETMODE, &vi.vi_mode) != -1 &&
+        ioctl(tty, FBIO_MODEINFO, &vi) != -1) {
+      cwidth = vi.vi_cwidth;
+      cheight = vi.vi_cheight;
+    }
+    signal(SIGUSR2, SIG_IGN);
+    mi.operation = MOUSE_MODE;
+    mi.u.mode.mode = 0;
+    mi.u.mode.signal = SIGUSR2;
+    if (ioctl(tty, CONS_MOUSECTL, &mi) != -1) {
+      signal(SIGUSR2, sysmouse);
+      mi.operation = MOUSE_SHOW;
+      ioctl(tty, CONS_MOUSECTL, &mi);
+      sysm_handler = sysm_process_mouse;
+    }
   }
-#endif
+#endif /* USE_SYSMOUSE */
   mouseActive = 1;
 }
 
@@ -1497,7 +1614,16 @@ void mouse_end()
 #ifdef USE_GPM
   else
     Gpm_Close();
-#endif
+#endif /* USE_GPM */
+#ifdef USE_SYSMOUSE
+  else {
+    mouse_info_t mi;
+    mi.operation = MOUSE_MODE;
+    mi.u.mode.mode = 0;
+    mi.u.mode.signal = 0;
+    ioctl(tty, CONS_MOUSECTL, &mi);
+  }
+#endif /* USE_SYSMOUSE */
   mouseActive = 0;
 }
 
@@ -1515,7 +1641,7 @@ void mouse_inactive()
     XTERM_OFF;
 }
   
-#endif
+#endif /* MOUSE */
 
 void flush_tty()
 {
