@@ -20,4 +20,4 @@ extern void appendText ( TextList *tl, char *s );
 extern char * popText ( TextList *tl );
 extern TextList *appendTextList(TextList *, TextList *);
 
-#endif
+#endif /* not TEXTLIST_H */
