@@ -6,12 +6,23 @@
 
 #include <signal.h>
 #include <setjmp.h>
+#include <errno.h>
+
+#include <sys/stat.h>
+#ifdef __EMX__
+#include <io.h>
+#include <strings.h>
+#endif /* __EMX__ */
 
 #include "html.h"
 #include "Str.h"
 #include "myctype.h"
 #include "regex.h"
 
+#ifdef	__WATT32__
+#define	write(a,b,c)	write_s(a,b,c)
+#endif /* __WATT32__ */
+
 #define NOPROXY_NETADDR  /* allow IP address for no_proxy */
 
 #ifdef INET6
@@ -25,6 +36,7 @@ int ai_family_order_table[3][3] = {
 
 static JMP_BUF  AbortLoading;
 
+/* XXX: note html.h SCM_* */
 static int
                 DefaultPort[] = {
   80,				/* http */
@@ -32,12 +44,14 @@ static int
   21,				/* ftp */
   21,				/* ftpdir */
   0,				/* local - not defined */
+  0,				/* ??? - not defined? */
+  0,				/* exec - not defined? */
   119,				/* nntp */
   119,				/* news */
   0,				/* mailto - not defined */
 #ifdef USE_SSL
   443,                          /* https */
-#endif
+#endif /* USE_SSL */
 };
 
 struct cmdtable schemetable[] = {
@@ -46,12 +60,13 @@ struct cmdtable schemetable[] = {
   {"ftp", SCM_FTP},
   {"local", SCM_LOCAL},
   {"file", SCM_LOCAL},
+  {"exec", SCM_LOCAL_CGI}, /* ??? */
   {"nntp", SCM_NNTP},
   {"news", SCM_NEWS},
   {"mailto", SCM_MAILTO},
 #ifdef USE_SSL
   {"https", SCM_HTTPS},
-#endif
+#endif /* USE_SSL */
   {NULL, SCM_UNKNOWN},
 };
 
@@ -76,6 +91,7 @@ static struct table2 DefaultGuess[] = {
   {"AU", "audio/basic"},
   {"gz", "application/x-gzip"},
   {"Z", "application/x-compress"},
+  {"bz2", "application/x-bzip"},
   {"tar", "application/x-tar"},
   {"zip", "application/x-zip"},
   {"lha", "application/x-lha"},
@@ -91,6 +107,23 @@ static struct table2 DefaultGuess[] = {
 
 #ifndef HTTP_DEFAULT_FILE
 #define HTTP_DEFAULT_FILE "/"
+#endif /* not HTTP_DEFAULT_FILE */
+
+#ifdef SOCK_DEBUG
+#include <stdarg.h>
+
+static void
+sock_log(char *message,...)
+{
+  FILE *f = fopen("zzzsocklog","a");
+  va_list va;
+
+  if (f == NULL) return;
+  va_start(va,message);
+  vfprintf(f,message,va);
+  fclose(f);
+}
+
 #endif
 
 static char *
@@ -100,11 +133,12 @@ DefaultFile(int scheme)
   case SCM_HTTP:
 #ifdef USE_SSL
   case SCM_HTTPS:
-#endif
+#endif /* USE_SSL */
     return allocStr(HTTP_DEFAULT_FILE, 0);
   case SCM_GOPHER:
     return allocStr("1", 0);
   case SCM_LOCAL:
+  case SCM_LOCAL_CGI:
   case SCM_FTP:
     return allocStr("/", 0);
   }
@@ -135,12 +169,12 @@ openSSLHandle(int sock)
 #if SSLEAY_VERSION_NUMBER < 0x0800
     ssl_ctx = SSL_CTX_new();
     X509_set_default_verify_paths(ssl_ctx->cert);
-#else
+#else /* SSLEAY_VERSION_NUMBER >= 0x0800 */
     SSLeay_add_ssl_algorithms();
     ssl_ctx = SSL_CTX_new(SSLv23_client_method());
     SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL);
     SSL_CTX_set_default_verify_paths(ssl_ctx);
-#endif
+#endif /* SSLEAY_VERSION_NUMBER >= 0x0800 */
     atexit(free_ssl_ctx);
   }
   handle = SSL_new(ssl_ctx);
@@ -166,7 +200,7 @@ SSL_write_from_file(SSL *ssl, char *file)
   }
 }
 
-#endif
+#endif /* USE_SSL */
 
 static void
 write_from_file(int sock, char *file)
@@ -209,24 +243,31 @@ openSocket(char *hostname,
   int *af;
   struct addrinfo hints, *res0, *res;
   int error;
-#else
+#else /* not INET6 */
   struct sockaddr_in hostaddr;
   struct hostent *entry;
-  struct servent *serv;
   struct protoent *proto;
+  unsigned short  s_port;
   int             a1, a2, a3, a4;
   unsigned long   adr;
-#endif
+#endif /* not INET6 */
   MySignalHandler(*trap) ();
 
   if (SETJMP(AbortLoading) != 0) {
+#ifdef SOCK_DEBUG
+    sock_log("openSocket() failed. reason: user abort\n");
+#endif
     goto error;
   }
   trap = signal(SIGINT, KeyAbort);
   if (fmInitialized)
     term_cbreak();
-  if (hostname == NULL)
+  if (hostname == NULL) {
+#ifdef SOCK_DEBUG
+      sock_log("openSocket() failed. reason: Bad hostname \"%s\"\n",hostname);
+#endif
     goto error;
+  }
 
 #ifdef INET6
   for (af = ai_family_order_table[DNS_order]; ; af++) {
@@ -234,9 +275,8 @@ openSocket(char *hostname,
       hints.ai_family = *af;
       hints.ai_socktype = SOCK_STREAM;
       if (remoteport_num != 0) {
-	  char portbuf[10];
-	  snprintf(portbuf, sizeof(portbuf), "%d", remoteport_num);
-	  error = getaddrinfo(hostname, portbuf, &hints, &res0);
+	  Str portbuf = Sprintf("%d", remoteport_num);
+	  error = getaddrinfo(hostname, portbuf->ptr, &hints, &res0);
       } else {
 	  error = -1;
       }
@@ -275,11 +315,8 @@ openSocket(char *hostname,
       freeaddrinfo(res0);
       break;
   }
-#else
-  if ((serv = getservbyname(remoteport_name, "tcp")) == NULL) {
-    serv = New(struct servent);
-  }
-  serv->s_port = htons(remoteport_num);
+#else /* not INET6 */
+  s_port = htons(remoteport_num);
   bzero((char *) &hostaddr, sizeof(struct sockaddr_in));
   if ((proto = getprotobyname("tcp")) == NULL) {
     /* protocol number of TCP is 6 */
@@ -287,6 +324,9 @@ openSocket(char *hostname,
     proto->p_proto = 6;
   }
   if ((sock = socket(AF_INET, SOCK_STREAM, proto->p_proto)) < 0) {
+#ifdef SOCK_DEBUG
+    sock_log("openSocket: socket() failed. reason: %s\n",strerror(errno));
+#endif
     goto error;
   }
   regexCompile("[0-9][0-9]*\\.[0-9][0-9]*\\.[0-9][0-9]*\\.[0-9][0-9]*",0);
@@ -295,33 +335,56 @@ openSocket(char *hostname,
     adr = htonl((a1<<24)|(a2<<16)|(a3<<8)|a4);
     bcopy((void*)&adr,(void*)&hostaddr.sin_addr,sizeof(long));
     hostaddr.sin_family = AF_INET;
-    hostaddr.sin_port = serv->s_port;
+    hostaddr.sin_port = s_port;
     if (connect(sock, (struct sockaddr *) & hostaddr,
                 sizeof(struct sockaddr_in)) < 0) {
+#ifdef SOCK_DEBUG
+      sock_log("openSocket: connect() failed. reason: %s\n",strerror(errno));
+#endif
       goto error;
     }
   }
   else {
     char **h_addr_list ;
     int result ;
+    message(Sprintf("Performing hostname lookup on %s\n", hostname)->ptr, 0, 0);
+    refresh();
     if ((entry = gethostbyname(hostname)) == NULL) {
+#ifdef SOCK_DEBUG
+      sock_log("openSocket: gethostbyname() failed. reason: %s\n",strerror(errno));
+#endif
       goto error;
     }
     hostaddr.sin_family = AF_INET;
-    hostaddr.sin_port = serv->s_port;
+    hostaddr.sin_port = s_port;
     for ( h_addr_list = entry->h_addr_list ; *h_addr_list ; h_addr_list ++ )
     {
       bcopy((void*)h_addr_list[0], (void*)&hostaddr.sin_addr, entry->h_length);
-
+#ifdef SOCK_DEBUG
+      adr = ntohl(*(long*)&hostaddr.sin_addr);
+      sock_log("openSocket: connecting %d.%d.%d.%d\n",
+	       (adr>>24)&0xff,
+	       (adr>>16)&0xff,
+	       (adr>>8)&0xff,
+	       adr&0xff);
+#endif
+      message(Sprintf("Connecting to %s\n", hostname)->ptr, 0, 0);
+      refresh();
       if ((result=connect(sock, (struct sockaddr *) & hostaddr,
                           sizeof(struct sockaddr_in))) == 0) {
         break;
       }
-    }
-    if ( result < 0 )
-      goto error;
-  }
+#ifdef SOCK_DEBUG
+      else {
+	sock_log("openSocket: connect() failed. reason: %s\n",strerror(errno));
+      }
 #endif
+    }
+    if ( result < 0 ) {
+      goto error;
+    }
+  }
+#endif /* not INET6 */
 
   if (fmInitialized)
     term_raw();
@@ -331,6 +394,7 @@ error:
   if (fmInitialized)
     term_raw();
   signal(SIGINT, trap);
+  
   return -1;
 
 }
@@ -358,6 +422,13 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
       copyParsedURL(p_url,current);
     goto do_label;
   }
+#ifdef __EMX__
+  if(isalpha(*url)&&url[1]==':'){	/* DRIVE:foo */
+    p_url->scheme=SCM_LOCAL;
+    p_url->file=allocStr(url,0);
+    goto analyze_file;
+  }
+#endif
   /* search for scheme */
   while (*p && (isalpha(*p) || *p == '.' || *p == '+' || *p == '-'))
     p++;
@@ -389,13 +460,18 @@ parseURL(char *url, ParsedURL * p_url, ParsedURL * current)
     goto analyze_file;
   }
   if (p_url->scheme == SCM_LOCAL) {	/*   file://foo		*/
+#ifdef __EMX__
+    p += 2;
+    goto analyze_file;
+#else
     if (p[2] == '/' || p[2] == '~') {	/*   file:///foo	*/
       p += 2;				/*   file://~user	*/
       goto analyze_file;
     }
+#endif
 #ifdef CYGWIN
     goto analyze_file;			/*   file://DRIVE/foo or file://machine/share/foo	*/
-#endif
+#endif /* CYGWIN */
   }
   p += 2;				/* scheme://foo		*/
 analyze_url:
@@ -572,7 +648,11 @@ parseURL2(char *url, ParsedURL *pu, ParsedURL *current)
     if (pu->file) {
       if (pu->scheme != SCM_GOPHER &&
 	  pu->scheme != SCM_NEWS && 
-	  pu->file[0] != '/') {
+	  pu->file[0] != '/'
+#ifdef __EMX__
+	  &&!(pu->scheme==SCM_LOCAL&&isalpha(pu->file[0])&&pu->file[1]==':')
+#endif
+	 ) {
         p = pu->file;
 	if (current->file) {
 	  tmp = Strnew_charp(current->file);
@@ -598,13 +678,32 @@ parseURL2(char *url, ParsedURL *pu, ParsedURL *current)
     }
   }
   if (pu->file) {
+#ifdef __EMX__
+    if(pu->scheme==SCM_LOCAL){
+      if (strncmp(pu->file,"/$LIB/",6)){
+	char*arg,abs[_MAX_PATH],tmp[_MAX_PATH];
+
+	if(!(arg=strchr(strcpy(tmp,pu->file),'?'))){
+	  _abspath(abs,tmp,_MAX_PATH);
+	  pu->file=cleanupName(abs);
+	}else{
+	  *arg=0;
+	  _abspath(abs,tmp,_MAX_PATH);
+	  *arg='?';
+	  pu->file=cleanupName(strcat(abs,arg));
+	}
+      }
+    }
+#else
     if (pu->scheme == SCM_LOCAL && pu->file[0] != '/') {
       tmp = Strnew_charp(currentdir());
       if (Strlastchar(tmp) != '/')
 	Strcat_char(tmp,'/');
       Strcat_charp(tmp,pu->file);
       pu->file = cleanupName(tmp->ptr);
-    } else if (pu->scheme != SCM_GOPHER && pu->scheme != SCM_NEWS &&
+    }
+#endif
+    else if (pu->scheme != SCM_GOPHER && pu->scheme != SCM_NEWS &&
 	       pu->file[0] == '/') {
       pu->file = cleanupName(pu->file);
     }
@@ -616,10 +715,10 @@ parsedURL2Str(ParsedURL *pu)
 {
   Str tmp = Strnew();
   static char *scheme_str[] = {
-    "http","gopher","ftp","ftp","file","nntp","news","mailto",
+    "http","gopher","ftp","ftp","file","file","exec","nntp","news","mailto",
 #ifdef USE_SSL
     "https",
-#endif
+#endif /* USE_SSL */
     };
 
   if (pu->scheme == SCM_UNKNOWN || pu->scheme == SCM_MISSING) {
@@ -655,7 +754,11 @@ parsedURL2Str(ParsedURL *pu)
       Strcat(tmp,Sprintf("%d",pu->port));
     }
   }
+#ifdef __EMX__
+  if (pu->file == NULL || ( pu->file[0] != '/' && pu->file[1] !=':' ) )
+#else
   if (pu->file == NULL || pu->file[0] != '/')
+#endif
     Strcat_char(tmp,'/');
   Strcat_charp(tmp,pu->file);
   if (pu->label) {
@@ -685,9 +788,9 @@ otherinfo(ParsedURL *target, ParsedURL *current, char *referer)
   else {
 #if LANG == JA
     Strcat_charp(s,"ja; q=1.0, en; q=0.5\r\n");
-#else  /* must be EN */
+#else  /* LANG != JA (must be EN) */
     Strcat_charp(s,"en; q=1.0\r\n");
-#endif
+#endif /* LANG != JA */
   }
   if (target->host) {
     Strcat_charp(s,"Host: ");
@@ -700,20 +803,20 @@ otherinfo(ParsedURL *target, ParsedURL *current, char *referer)
     Strcat_charp(s, "Pragma: no-cache\r\n");
     Strcat_charp(s, "Cache-control: no-cache\r\n");
   }
- if(!NoSendReferer){
-  if (referer == NULL && current && current->scheme != SCM_LOCAL &&
-      (current->scheme != SCM_FTP ||
-       (current->user == NULL && current->pass == NULL))) {
-    Strcat_charp(s,"Referer: ");
-    Strcat(s,parsedURL2Str(current));
-    Strcat_charp(s,"\r\n");
+  if(!NoSendReferer){
+    if (referer == NULL && current && current->scheme != SCM_LOCAL &&
+	(current->scheme != SCM_FTP ||
+	 (current->user == NULL && current->pass == NULL))) {
+      Strcat_charp(s,"Referer: ");
+      Strcat(s,parsedURL2Str(current));
+      Strcat_charp(s,"\r\n");
+    }
+    else if (referer != NULL && referer != NO_REFERER) {
+      Strcat_charp(s,"Referer: ");
+      Strcat_charp(s,referer);
+      Strcat_charp(s,"\r\n");
+    }
   }
-  else if (referer != NULL && referer != NO_REFERER) {
-    Strcat_charp(s,"Referer: ");
-    Strcat_charp(s,referer);
-    Strcat_charp(s,"\r\n");
-  }
- }
   return s->ptr;
 }
 
@@ -724,7 +827,7 @@ HTTPrequest(ParsedURL *pu, ParsedURL *current, HRequest *hr, TextList *extra)
   TextListItem *i;
 #ifdef USE_COOKIE
   Str cookie;
-#endif
+#endif /* USE_COOKIE */
   switch (hr->command) {
   case HR_COMMAND_CONNECT:
     tmp = Strnew_charp("CONNECT ");
@@ -764,7 +867,7 @@ HTTPrequest(ParsedURL *pu, ParsedURL *current, HRequest *hr, TextList *extra)
     Strcat(tmp, cookie);  
     Strcat_charp(tmp,"\r\n");
   }
-#endif
+#endif /* USE_COOKIE */
   if (hr->command == HR_COMMAND_POST) {
     if (hr->request->enctype == FORM_ENCTYPE_MULTIPART){
       Strcat_charp(tmp,"Content-type: multipart/form-data; boundary=");
@@ -784,7 +887,7 @@ HTTPrequest(ParsedURL *pu, ParsedURL *current, HRequest *hr, TextList *extra)
     Strcat_charp(tmp,"\r\n");
 #ifdef DEBUG
   fprintf(stderr, "HTTPrequest: [ %s ]\n\n", tmp->ptr);
-#endif
+#endif /* DEBUG */
   return tmp;
 }
 
@@ -802,7 +905,7 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
   HRequest        hr;
 #ifdef USE_SSL
   SSL            *sslh;
-#endif
+#endif /* USE_SSL */
 
   if (ouf) {
     uf = *ouf;
@@ -814,7 +917,7 @@ openURL(char *url, ParsedURL * pu, ParsedURL * current,
     uf.close = (void(*)())fclose;
     uf.encoding = ENC_7BIT;
     uf.is_cgi = NIL;
-    uf.gzip = 0;
+    uf.compression = 0;
   }
 
 retry:
@@ -829,6 +932,9 @@ retry:
     }
     else {
       /* given URL must be null string */
+#ifdef SOCK_DEBUG
+      sock_log("given URL must be null string\n");
+#endif
       return uf;
     }
   }
@@ -854,12 +960,14 @@ retry:
   
   switch (pu->scheme) {
   case SCM_LOCAL:
+  case SCM_LOCAL_CGI:
     if (request && request->body) {
       /* local CGI: POST */
       uf.stream.f = localcgi_post(pu->file,request);
       if (uf.stream.f == NULL) goto ordinary_local_file;
       uf.close = (void(*)())pclose;	
       uf.is_cgi = T;
+      uf.scheme = SCM_LOCAL_CGI;
     }	
     else if ((p = strchr(pu->file,'?')) != NULL) {
       /* lodal CGI: GET */
@@ -872,29 +980,50 @@ retry:
       }
       uf.close = (void(*)())pclose; 
       uf.is_cgi = T;
+      uf.scheme = SCM_LOCAL_CGI;
     }      
-    else if (!strncmp(pu->file+strlen(pu->file)-4,".cgi",4)) {
+#ifdef __EMX__
+    else if (!strncmp(pu->file+strlen(pu->file)-4,".cmd",4))
+#else
+    else if (!strncmp(pu->file+strlen(pu->file)-4,".cgi",4))
+#endif
+    {
       /* lodal CGI: GET */
       uf.stream.f = localcgi_get(pu->file,"");
       if (uf.stream.f == NULL) goto ordinary_local_file;
       uf.close = (void(*)())pclose; 
       uf.is_cgi = T;
+      uf.scheme = SCM_LOCAL_CGI;
     }      
     else {
     ordinary_local_file:
       uf.stream.f = examineFile(pu->file,&uf.close);
     }
-    if (uf.stream.f == NULL && document_root != NULL) {
-      Strcat_charp(tmp,document_root);
-      if (Strlastchar(tmp) != '/')
-	Strcat_char(tmp,'/');
-      Strcat_charp(tmp,pu->file);
-      uf.stream.f = examineFile(tmp->ptr,&uf.close);
+    if (uf.stream.f == NULL) {
+      if (dir_exist(pu->file))
+	return uf;
+      if (document_root != NULL) {
+	Strcat_charp(tmp,document_root);
+	if (Strlastchar(tmp) != '/')
+	  Strcat_char(tmp,'/');
+	Strcat_charp(tmp,pu->file);
+	uf.stream.f = examineFile(tmp->ptr,&uf.close);
+      }
     }
-    if (uf.stream.f == NULL && strncasecmp(url,"http://",7) != 0) {
-      /* retry it as http:// */
-      url = Strnew_m_charp("http://",url,NULL)->ptr;
-      goto retry;
+    if (uf.stream.f == NULL){
+      if (dir_exist(pu->file)) {
+      	/* If it is a directry	*/
+	/* don't assume http:// */
+#ifdef SOCK_DEBUG
+        sock_log("%s is a directory and can't open\n",pu->file);
+#endif
+	return uf;
+      }
+      if (strncasecmp(url,"http://",7) != 0) {
+	/* retry it as http:// */
+        url = Strnew_m_charp("http://",url,NULL)->ptr;
+        goto retry;
+      }
     }
     return uf;
   case SCM_FTP:
@@ -907,7 +1036,7 @@ retry:
 			schemetable[FTP_proxy_parsed.scheme].cmdname,
 			FTP_proxy_parsed.port);
       if (sock < 0)
-	goto no_ftp_proxy;
+        return uf;
       tmp = HTTPrequest(pu,current,&hr,NULL);
       write(sock, tmp->ptr, tmp->length);
     } else {
@@ -921,7 +1050,7 @@ retry:
   case SCM_HTTP:
 #ifdef USE_SSL
   case SCM_HTTPS:
-#endif
+#endif /* USE_SSL */
     if (pu->file == NULL)
       pu->file = allocStr("/",0);
     if (request && request->method == FORM_METHOD_POST && request->body)
@@ -950,8 +1079,12 @@ retry:
 			schemetable[HTTP_proxy_parsed.scheme].cmdname,
 			HTTP_proxy_parsed.port);
 #endif
-      if (sock < 0)
-	goto no_http_proxy;
+      if (sock < 0) {
+#ifdef SOCK_DEBUG
+	sock_log("Can't open socket\n");
+#endif
+        return uf;
+      }
       save_label = pu->label;
       pu->label = NULL;
 #ifdef USE_SSL
@@ -966,7 +1099,7 @@ retry:
 	  *status = HTST_NORMAL;
 	}
       } else
-#endif
+#endif /* USE_SSL */
       {
         tmp = HTTPrequest(pu,current,&hr,extra_header);
         *status = HTST_NORMAL;
@@ -988,7 +1121,7 @@ retry:
 	  return uf;
 	}
       }
-#endif
+#endif /* USE_SSL */
       hr.flag |= HR_FLAG_LOCAL;
       tmp = HTTPrequest(pu,current,&hr,extra_header);
       *status = HTST_NORMAL;
@@ -1010,7 +1143,7 @@ retry:
       }
       return uf;
     } else
-#endif
+#endif /* USE_SSL */
     {
       write(sock, tmp->ptr, tmp->length);
 #ifdef HTTP_DEBUG
@@ -1019,7 +1152,7 @@ retry:
 	fwrite(tmp->ptr,sizeof(char),tmp->length,ff);
 	fclose(ff);
       }
-#endif
+#endif /* HTTP_DEBUG */
       if (hr.command == HR_COMMAND_POST &&
 	  request->enctype == FORM_ENCTYPE_MULTIPART)
 	write_from_file(sock, request->body);
@@ -1033,7 +1166,7 @@ retry:
 			schemetable[GOPHER_proxy_parsed.scheme].cmdname,
 			GOPHER_proxy_parsed.port);
       if (sock < 0)
-	goto no_gopher_proxy;
+        return uf;
       tmp = HTTPrequest(pu,current,&hr,NULL);
     } else {
   no_gopher_proxy:
@@ -1056,8 +1189,8 @@ retry:
     sock = openSocket(p, "nntp", pu->port);
     if (sock < 0)
       return uf;
-    f = fdopen(sock, "r");
-    fw = fdopen(sock, "w");
+    f = fdopen(sock, "rb");
+    fw = fdopen(sock, "wb");
     if (f == NULL || fw == NULL)
       return uf;
     tmp = Strfgets(f);
@@ -1084,7 +1217,8 @@ nntp_error:
   case SCM_UNKNOWN:
     return uf;
   }
-  uf.stream.f = fdopen(sock, "r");
+  uf.stream.f = fdopen(sock, "rb");
+  uf.stream_type = SMT_FILE;
   uf.close = (void(*)())fclose;
   return uf;
 }
@@ -1093,6 +1227,8 @@ char           *
 guessContentTypeFromTable(struct table2 * table, char *filename)
 {
   char           *p;
+  if (table == NULL)
+    return NULL;
   p = &filename[strlen(filename) - 1];
   while (filename < p && *p != '.')
     p--;
@@ -1110,8 +1246,14 @@ guessContentTypeFromTable(struct table2 * table, char *filename)
 char           *
 guessContentType(char *filename)
 {
+  char *ret;
+
   if (filename == NULL)
     return NULL;
+  if ((ret = guessContentTypeFromTable(UserMimeTypes, filename)) != NULL)
+    return ret;
+  if ((ret = guessContentTypeFromTable(SysMimeTypes, filename)) != NULL)
+    return ret;
   return guessContentTypeFromTable(DefaultGuess, filename);
 }
 
